# =========================================================================== #
import secrets
from http import HTTPMethod
from typing import Any, Callable, Dict, Generic, Set, Tuple, TypeVar, overload

from fastapi import HTTPException
from sqlalchemy.orm import Session

# --------------------------------------------------------------------------- #
from captura.auth import Token
from captura.controllers.access import Access, H, with_access
from captura.controllers.base import (
    Data,
    DataResolvedGrant,
    KindData,
    ResolvedAssignmentCollection,
    ResolvedAssignmentDocument,
    ResolvedCollection,
    ResolvedDocument,
    ResolvedEvent,
    ResolvedGrantDocument,
    ResolvedGrantUser,
    ResolvedUser,
)
from captura.controllers.delete import (
    AssocData,
    DataResolvedAssignment,
    Delete,
    WithDelete,
)
from captura.err import ErrAssocRequestMustForce
from captura.models import (
    Assignment,
    Collection,
    Document,
    Event,
    Grant,
    KindEvent,
    KindObject,
    Level,
    PendingFrom,
    User,
)
from captura.schemas import (
    CollectionCreateSchema,
    CollectionUpdateSchema,
    DocumentCreateSchema,
    DocumentUpdateSchema,
    GrantCreateSchema,
    UserCreateSchema,
    UserUpdateSchema,
    mwargs,
)

# Typehints for assoc callback.
# NOTE: Tried protocol, too much of a pain in the ass.


AssocCallbackGrant = Callable[
    [
        Data[ResolvedGrantDocument] | Data[ResolvedGrantUser],
        User | Document,
    ],
    Grant,
]
AssocCallbackAssignment = Callable[
    [
        Data[ResolvedAssignmentDocument] | Data[ResolvedAssignmentCollection],
        Collection | Document,
    ],
    Assignment,
]
AssocCallback = AssocCallbackGrant | AssocCallbackAssignment

# --------------------------------------------------------------------------- #


# T_with_empty_data_self = TypeVar(
#     "T_with_empty_data_self",
#     bound=WithDelete,
# )
#

# NOTE: Apply to methods like :func:`with_access`. In this case access is only
#       checked by verifying that the user has a valid token.
#       The goal is to create empty data and use that as the entry point
#       instead of a method of :class:`Access`
# def with_empty_data(
#     meth: Callable[[T_with_empty_data_self, Data[T_Data]], Data[T_Data]],
# ) -> Callable[[T_with_empty_data_self, Data[T_Data] | None], Data[T_Data]]:
#
#     def reject_nonempty(
#         self: T_with_empty_data_self, data: Data[T_Data] | None = None
#     ) -> Data[T_Data]:
#         data: Data[T_Data]
#         if data is None:
#             data = Data[T_Data].model_vali
#
#         data_empty = data
#
#         return meth(self, data)
#
#     reject_nonempty.__name__ = meth.__name__
#     reject_nonempty.__doc__ = meth.__doc__
#
#     return reject_nonempty


# --------------------------------------------------------------------------- #
# Generic upsert.

T_Create = TypeVar(
    "T_Create",
    CollectionCreateSchema,
    DocumentCreateSchema,
    UserCreateSchema,
    GrantCreateSchema,
)


class Create(WithDelete, Generic[T_Create]):
    # NOTE: `PUT` will only be supported on assignments and grants for now
    #       (force overwriting of existing). This is because `PATCH` will be
    #       used to accept grants by grant uuid.

    _create_data: T_Create | None

    def __init__(
        self,
        session: Session,
        token: Token | Dict[str, Any] | None,
        method: HTTPMethod | str,
        *,
        api_origin: str,
        force: bool = False,
        access: Access | None = None,
        delete: Delete | None = None,
        create_data: T_Create | None = None,
    ):
        if method != H.POST and method != H.PUT:
            msg = f"Cannot accept method `{method}` (expected one of `PUT` or "
            raise ValueError(msg + "`PATCH`.")

        super().__init__(
            session,
            token,
            method,
            api_origin=api_origin,
            force=force,
            access=access,
            delete=delete,
        )
        self._create_data = create_data

    @property
    def create_data(self) -> T_Create:
        if (create_data := self._create_data) is None:
            raise AttributeError("`create_data` is not yet set.")
        return create_data

    @create_data.setter
    def create_data(self, v: T_Create):
        self._create_data = v

    # ----------------------------------------------------------------------- #
    # Helpers

    @property
    def event_common(self) -> Dict[str, Any]:
        kind = KindEvent.create if self.method == H.POST else KindEvent.update
        return dict(**super().event_common, kind=kind)

    # TODO: Make this work with generics.
    def assoc(
        self,
        data: DataResolvedGrant | DataResolvedAssignment,
        create_assoc: AssocCallback,
    ) -> Tuple[DataResolvedGrant | DataResolvedAssignment, AssocData]:
        """Symetrically handle forced creation.

        When :attr:`force` is ``True``, lingering assignments staged for
        deletion will be cleaned up by :attr:`delete`. Otherwise, if
        :param:`data` specified exising assignments, an error will be raised.

        :param assoc_args: Common arguments for creating the associations. This
            should be everything that is not specified by :param:`data` or
            :attr:`create_data`. For instance, grant should pass key value
            pairs for ``pending_from`` and ``uuid_parent``. This parameter
            should be used for internal values and not user inputs, which
            should be added to `create_data`.
        :returns: The active target ids that do not yet have an assignment to
            their source. The requirement that they be active means that doing
            the same `POST` twice should be indempotent.
        """

        # NOTE: By default, only create for those that hove no assocs.
        assoc_data, model_assoc = self.delete.split_assocs(data)
        uuid_target_create = assoc_data.uuid_target_none.copy()

        if self.method != HTTPMethod.POST:
            raise HTTPException(405)

        event_rm: Event | None = None
        if not self.force:
            if assoc_data.uuid_assoc_deleted:
                raise ErrAssocRequestMustForce.httpexception(
                    "_msg_force",
                    400,
                    kind_target=data.data.kind_target,
                    kind_source=data.data.kind_source,
                    kind_assoc=data.data.kind_assoc,
                    uuid_target=list(assoc_data.uuid_target_deleted),
                    uuid_source=data.data.uuid_source,
                    uuid_assoc=list(assoc_data.uuid_assoc_deleted),
                )
        else:
            q_del, uuid_assoc_rm = assoc_data.q_del(model_assoc, True, rm_active=False)
            q_assocs = model_assoc.q_uuid(uuid_assoc_rm)
            assocs_rm = tuple(self.session.scalars(q_assocs))

            self.session.execute(q_del)
            event_rm = self.delete.create_event_assoc(data, assocs_rm)
            uuid_target_create |= assoc_data.uuid_target_deleted

        targets = tuple(tt for tt in data.data.target if tt.uuid in uuid_target_create)
        assocs: Dict[str, Grant] | Dict[str, Assignment] = {
            target.uuid: create_assoc(data, target)  # type: ignore
            for target in targets
        }

        # WARNING! OVERWRITES!
        resolved_raw = {
            data.data._attr_name_source: data.data.source,
            data.data._attr_name_target: targets,
            data.data._attr_name_assoc: assocs,
        }
        if data.data.kind_assoc == KindObject.grant:
            resolved_raw["token_user_grants"] = getattr(data.data, "token_user_grants")

        data_final: DataResolvedGrant | DataResolvedAssignment = mwargs(
            Data[data.data.__class__],  # type: ignore
            data=data.data.__class__.model_validate(resolved_raw),
        )

        event_create = self.create_event_assoc(data_final, assocs)
        if event_rm is not None:
            event = Event(
                **self.event_common,
                uuid_obj=data_final.data.uuid_source,
                kind_obj=data_final.data.kind_source,
                children=[event_rm, event_create],
            )
            event.kind = KindEvent.upsert
        else:
            event = event_create

        data_final.event = event
        setattr(data_final.data, data_final.data._attr_name_assoc, assocs)

        # ------------------------------------------------------------------- #
        #
        # NOTE: This snippet was helpful in debugging, do not delete.
        # import json
        # from pydantic import TypeAdapter
        # from rich.table import Table
        # from captura.schemas import AssignmentSchema, GrantSchema
        #
        # table = Table(show_lines=False, show_header=False)
        # table.add_column()
        # table.add_column()
        # table.add_row("data.data.assoc", str(data.data.assoc))
        # table.add_row("self.force", str(self.force))
        # table.add_row("assoc_data", assoc_data.render())
        # self.session.add_all(assocs.values())
        # table.add_row(
        #     "assocs",
        #     str([[aa.uuid_collection, aa.uuid_document] for aa in assocs.values()]),
        # )
        # table.title = "From Create"
        #
        # ------------------------------------------------------------------- #

        return data_final, assoc_data

    def create_event_assoc(
        self,
        data: DataResolvedGrant | DataResolvedAssignment,
        grants: Dict[str, Grant] | Dict[str, Assignment],
    ) -> Event:
        return Event(
            **(event_common := self.event_common),
            uuid=secrets.token_urlsafe(8),
            uuid_obj=data.data.uuid_source,
            kind_obj=data.data.kind_source,
            children=[
                Event(
                    **event_common,
                    uuid=secrets.token_urlsafe(8),
                    kind_obj=data.data.kind_target,
                    uuid_obj=uuid_target,
                    children=[
                        Event(
                            **event_common,
                            uuid=secrets.token_urlsafe(8),
                            kind_obj=data.data.kind_assoc,
                            uuid_obj=grant.uuid,
                        )
                    ],
                )
                for uuid_target, grant in grants.items()
            ],
        )

    # ----------------------------------------------------------------------- #
    # Assocs
    # @overload
    # def create_assignment(
    #     self,
    #     data: Data[ResolvedAssignmentCollection],
    #     target: Document,
    # ) -> Assignment: ...
    #
    # @overload
    # def create_assignment(
    #     self,
    #     data: Data[ResolvedAssignmentDocument],
    #     target: Collection,
    # ) -> Assignment: ...

    def create_assignment(
        self,
        data: Data[ResolvedAssignmentDocument] | Data[ResolvedAssignmentCollection],
        target: Collection | Document,
    ) -> Assignment:
        id_source_name = f"id_{data.data.kind_source.name}"
        id_target_name = f"id_{data.data.kind_target.name}"
        id_source_value = data.data.source.id  # type: ignore

        kwargs = {
            "uuid": secrets.token_urlsafe(8),
            id_source_name: id_source_value,
            id_target_name: target.id,
        }
        return Assignment(
            **kwargs,
            **self.create_data.model_dump(),
        )

    def assignment_document(
        self,
        data: Data[ResolvedAssignmentDocument],
    ) -> Data[ResolvedAssignmentDocument]:
        data, *_ = self.assoc(data, self.create_assignment)  # type: ignore[assignment]
        assert data.event
        return data

    def assignment_collection(
        self,
        data: Data[ResolvedAssignmentCollection],
    ) -> Data[ResolvedAssignmentCollection]:
        data, *_ = self.assoc(data, self.create_assignment)  # type: ignore[assignment]
        return data

    # @overload
    # def create_grant(
    #     self,
    #     data: Data[ResolvedGrantDocument],
    #     target: User,
    # ) -> Grant: ...
    #
    # @overload
    # def create_grant(
    #     self,
    #     data: Data[ResolvedGrantUser],
    #     target: Document,
    # ) -> Grant: ...

    def create_grant(
        self,
        data: Data[ResolvedGrantDocument] | Data[ResolvedGrantUser],
        target: Document | User,
    ) -> Grant:
        # NOTE: Grants should be indexed by uuids of documents.
        grant_parent_uuid: str | None
        match data.data.kind_source:
            case KindObject.document:
                # NOTE: Permission of granter since they are inviting a user in
                #       this case. This means that the invitor (a user who
                #       already owns the document) will be the one responsible
                #       for the grants creation.
                parent_grants = data.data.token_user_grants.values()
                if not (n := len(parent_grants)):
                    detail = dict(
                        msg="No parent grant.",
                        uuid_target=target.uuid,
                        uuid_source=data.data.source.uuid,
                        kind_target=data.data.kind_target,
                        kind_source=data.data.kind_source,
                    )
                    raise HTTPException(403, detail=detail)
                elif n > 1:
                    raise HTTPException(500, detail="Granter has too many grants.")

                (parent_grant,) = parent_grants
                grant_parent_uuid = parent_grant.uuid
                pending_from = PendingFrom.grantee
            case KindObject.user:
                grant_parent_uuid = None
                # ^^^^^^^^^^^^^^^^^ Assigned later by granter who accepts the
                #                   request. Grant of the granter to remove
                #                   this from pending state will be created
                #                   later.
                pending_from = PendingFrom.granter
            case bad:
                raise ValueError(f"Invalid source `{bad}`.")

        id_source_name = f"id_{data.data.kind_source.name}"
        id_target_name = f"id_{data.data.kind_target.name}"
        id_source_value = data.data.source.id  # type: ignore

        kwargs = {
            "uuid": secrets.token_urlsafe(8),
            "uuid_parent": grant_parent_uuid,
            "pending_from": pending_from,
            "pending": True,
            id_source_name: id_source_value,
            id_target_name: target.id,
        }
        return Grant(**kwargs, **self.create_data.model_dump())

    def grant_user(
        self,
        data: Data[ResolvedGrantUser],
    ) -> Data[ResolvedGrantUser]:
        data, *_ = self.assoc(data, self.create_grant)  # type: ignore[assignment]
        return data

    def grant_document(
        self,
        data: Data[ResolvedGrantDocument],
    ) -> Data[ResolvedGrantDocument]:
        data, *_ = self.assoc(data, self.create_grant)  # type: ignore[assignment]
        return data

    a_assignment_document = with_access(Access.assignment_document)(assignment_document)
    a_assignment_collection = with_access(Access.assignment_collection)(
        assignment_collection
    )
    a_grant_document = with_access(Access.grant_document)(grant_document)
    a_grant_user = with_access(Access.grant_user)(grant_user)

    # ----------------------------------------------------------------------- #
    # Others
    #
    # NOTE: Nested creation will not be put here. That will be done later after
    #       an initial release.

    def user(
        self,
        data: Data[ResolvedUser],
    ) -> Data[ResolvedUser]:
        user = User(
            **self.create_data.model_dump(),
            uuid=(uuid_user := secrets.token_urlsafe(8)),
        )
        data_create = data.empty(KindData.user)
        data_create.data.users = (user,)
        data_create.event = Event(
            **self.event_common,
            uuid_user=uuid_user,
            kind_obj=KindObject.user,
            uuid_obj=uuid_user,
            detail="User created.",
        )
        return data_create

    # NOTE: Ideally consuming functions would just pass in an empty data and
    #       this would fill it out. Unfortunately mutations make things more
    #       complicated so the behaviour here is not as such.
    def collection(
        self,
        data: Data[ResolvedCollection],
    ) -> Data[ResolvedCollection]:
        collection = Collection(
            **self.create_data.model_dump(),
            user=data.token_user,
            uuid=(uuid_collection := secrets.token_urlsafe(8)),
        )
        data_create = data.empty(KindData.collection)
        data_create.data.collections = (collection,)
        data_create.event = Event(
            **self.event_common,
            kind_obj=KindObject.collection,
            uuid_obj=uuid_collection,
            detail="Collection created.",
        )

        return data_create

    def document(
        self,
        data: Data[ResolvedDocument],
    ) -> Data[ResolvedDocument]:
        uuid_document, uuid_grant = (secrets.token_urlsafe(8) for _ in range(2))
        user = data.token_user or self.token_user
        data.data.documents = (
            document := Document(
                **self.create_data.model_dump(),
            ),
        )

        data.data.token_user_grants = {
            user.uuid: (
                Grant(
                    id_user=user.id,
                    id_document=document.id,
                    level=Level.own,
                    pending=False,
                    pending_from=PendingFrom.created,
                    uuid=uuid_grant,
                    # id_user_granter=user.id,
                )
            )
        }
        data.event = Event(
            **self.event_common,
            kind_obj=KindObject.document,
            uuid_obj=uuid_document,
            detail="Document created.",
            children=[
                Event(
                    **self.event_common,
                    kind_obj=KindObject.user,
                    uuid_obj=user.uuid,
                    detail="Ownership granted by creation.",
                    children=[
                        Event(
                            **self.event_common,
                            kind_obj=KindObject.grant,
                            uuid_obj=uuid_grant,
                            detail="Ownership created for user.",
                        )
                    ],
                )
            ],
        )
        return data

    # e_document = with_empty_data(document)


# --------------------------------------------------------------------------- #
# Update.

T_Update = TypeVar(
    "T_Update",
    CollectionUpdateSchema,
    DocumentUpdateSchema,
    UserUpdateSchema,
    bool,  # approve or reject grant
)


class Update(WithDelete, Generic[T_Update]):
    _update_data: T_Update | None

    def __init__(
        self,
        session: Session,
        token: Token | Dict[str, Any] | None,
        method: HTTPMethod | str,
        *,
        api_origin: str,
        force: bool = False,
        access: Access | None = None,
        delete: Delete | None = None,
        update_data: T_Update | None = None,
    ):
        if method != H.PATCH:
            msg = f"Cannot accept method `{method}` (expected one of `PUT` or "
            raise ValueError(msg + "`PATCH`).")

        super().__init__(
            session,
            token,
            method,
            api_origin=api_origin,
            force=force,
            access=access,
            delete=delete,
        )
        self._update_data = update_data

    @property
    def update_data(self) -> T_Update:
        if (v := self._update_data) is None:
            raise ValueError("`update_data` has not been set.")
        return v

    @update_data.setter
    def update_data(self, v) -> None:
        if self._update_data is not None:
            raise ValueError("Cannot overwrite existing update data.")
        self._update_data = v

    @property
    def event_common(self) -> Dict[str, Any]:
        v = super().event_common
        v.update(kind=KindEvent.update)
        return v

    @overload
    def generic_update(
        self,
        data: Data[ResolvedUser],
        exclude: Set[str] = set(),
        commit: bool = True,
    ) -> Tuple[Data[ResolvedUser], UserUpdateSchema]: ...

    @overload
    def generic_update(
        self,
        data: Data[ResolvedCollection],
        exclude: Set[str] = set(),
        commit: bool = True,
    ) -> Tuple[Data[ResolvedCollection], CollectionUpdateSchema]: ...

    @overload
    def generic_update(
        self,
        data: Data[ResolvedDocument],
        exclude: Set[str] = set(),
        commit: bool = True,
    ) -> Tuple[Data[ResolvedDocument], DocumentUpdateSchema]: ...

    # NOTE: Document updates are not generic
    def generic_update(
        self,
        data: Data[ResolvedUser] | Data[ResolvedCollection] | Data[ResolvedDocument],
        exclude: Set[str] = set(),
        commit: bool = True,
    ) -> (
        Tuple[Data[ResolvedUser], UserUpdateSchema]
        | Tuple[Data[ResolvedCollection], CollectionUpdateSchema]
        | Tuple[Data[ResolvedDocument], DocumentUpdateSchema]
    ):
        session = self.session

        item: User | Collection | Document
        match data.data:
            case object(users=(User() as item,)):  # type: ignore[misc]
                pass
            case object(collections=(Collection() as item,)):  # type: ignore[misc]
                pass
            case object(documents=(Document() as item,)):  # type: ignore[misc]
                pass
            case _:
                raise ValueError("Updating many is not supported.")

        param: T_Update = self.update_data
        match self.update_data.kind_mapped:
            case KindObject.user | KindObject.collection | KindObject.document:
                param_dict = param.model_dump(exclude_none=True, exclude=exclude)
            case bad:
                msg = f"Incorrect parameter type (got `{bad}` expected "
                msg += f"{KindObject.user}` or `{KindObject.document})."
                raise ValueError(msg)

        # NOTE: @optional will raise 422 if all `null`
        data.event = Event(
            uuid_obj=item.uuid,
            kind_obj=param.kind_mapped,
            detail="Fields updated using `generic_update`.",
            **self.event_common,
        )

        for column, value in param_dict.items():
            setattr(item, column, value)
            data.event.children.append(
                Event(
                    uuid_obj=item.uuid,
                    kind_obj=param.kind_mapped,
                    detail=f"Field `{column}` updated.",
                    **self.event_common,
                )
            )

        if commit:
            session.add(item)
            session.add(data.event)
            session.commit()
            session.refresh(data.event)
            session.refresh(item)

        return data, param  # type: ignore

    # ------------------------------------------------------------------------ #

    def user(self, data: Data[ResolvedUser]) -> Data[ResolvedUser]:
        session = self.session
        data, _ = self.generic_update(data, commit=True)
        user, *_ = data.data.users
        data.event = Event(
            **self.event_common,
            kind_obj=KindObject.collection,
            uuid_obj=user.uuid,
            detail="User updated.",
            children=[data.event],
        )
        session.add(data.event)
        session.add_all(data.data.users)
        session.commit()
        session.refresh(data.event)
        for item in data.data.users:
            session.refresh(item)

        return data

    a_user = with_access(Access.d_user)(user)

    # ------------------------------------------------------------------------ #

    def collection(self, data: Data[ResolvedCollection]) -> Data[ResolvedCollection]:
        param: CollectionUpdateSchema
        token_user = data.token_user or self.token_user
        data, param = self.generic_update(data, {"uuid_user"}, commit=False)

        collection, *_ = data.data.collections
        event_root = data.event
        data.event = Event(
            **self.event_common,
            kind_obj=KindObject.collection,
            uuid_obj=collection.uuid,
            detail="Collection updated.",
            children=[],
        )

        if event_root is not None and event_root.children:
            data.event.children.append(event_root)

        session = self.session
        if param.uuid_user is None:
            return data

        # Transfer ownership, same transaction so easy to rollback.
        collection = data.data.collections[0]
        user = User.if_exists(
            session,
            param.uuid_user,
        )
        collection.id_user = user.id
        data.event.children.append(
            Event(
                **self.event_common,
                kind_obj=KindObject.user,
                uuid_obj=token_user.uuid,
                detail="Ownership of collection transfered.",
            )
        )

        return data

    a_collection = with_access(Access.d_collection)(collection)

    # ------------------------------------------------------------------------ #

    def document(self, data: Data[ResolvedDocument]) -> Data[ResolvedDocument]:
        session = self.session
        # token_user = data.token_user or self.token_user

        param: DocumentUpdateSchema
        data, param = self.generic_update(data, {"content"}, commit=False)
        document = data.data.documents[0]

        if param.content is None:
            session.add(document)
            session.add(data.event)
            session.commit()
            session.refresh(data.event)
            return data

        document.content = param.content

        session.add(data.event)
        session.commit()
        session.refresh(data.event)
        return data

    a_document = with_access(Access.d_document)(document)

    # ------------------------------------------------------------------------ #

    def create_event_grant(
        self, data: DataResolvedGrant, grants_pending: Tuple[Grant, ...]
    ) -> Event:
        grant_uuid_target_key = f"uuid_{data.data.kind_target.name}"
        return Event(
            **(event_common := self.event_common),
            uuid=secrets.token_urlsafe(8),
            uuid_obj=data.data.uuid_source,
            kind_obj=data.data.kind_source,
            children=[
                Event(
                    **event_common,
                    uuid=secrets.token_urlsafe(8),
                    kind_obj=data.data.kind_target,
                    uuid_obj=getattr(grant, grant_uuid_target_key),
                    children=[
                        Event(
                            **event_common,
                            uuid=secrets.token_urlsafe(8),
                            kind_obj=data.data.kind_assoc,
                            uuid_obj=grant.uuid,
                        )
                    ],
                )
                for grant in grants_pending
            ],
        )

    def grant_user(
        self,
        data: Data[ResolvedGrantUser],
    ) -> Data[ResolvedGrantUser]:
        """Allow `token_user` to accept grants for many documents."""

        # NOTE: Pending invitations should not be included in data. This
        #       url should still require the user uuid in the url because admins
        #       will be able to modify permissions generally.
        token_user = data.token_user or self.token_user
        if token_user.uuid != data.data.user.uuid:
            detail = dict(
                msg="User cannot accept grants of other users.",
                uuid_user=data.data.uuid_user,
                uuid_user_token=token_user.uuid,
            )
            raise HTTPException(403, detail=detail)

        pending_grants = data.data.grants
        for grant in (pending_values := pending_grants.values()):
            # NOTE: Aleady checked by access. Here incase errs.
            if grant.pending_from != PendingFrom.grantee:
                raise HTTPException(
                    500,
                    detail="Cannot approve grant pending from `grantee` or `created`.",
                )

            # grant.uuid_parent = token_user_grant.uuid
            # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Assigned by granter on invite
            grant.pending = False
            self.session.merge(grant)

        event = self.create_event_grant(data, tuple(pending_values))
        event.detail = "Access granted."
        data.event = event
        return data

    def grant_document(
        self,
        data: Data[ResolvedGrantDocument],
    ) -> Data[ResolvedGrantDocument]:
        "Granter approves many users for owned doc."

        token_user = data.token_user or self.token_user
        token_user_grant = data.data.token_user_grants[token_user.uuid]

        # Double check the level. The first case should not really happen.
        if token_user.uuid != token_user_grant.uuid_user:
            detail = dict(
                uuid_user=token_user.uuid,
                uuid_user_token_from_grant=token_user_grant.uuid,
                msg="Token user grant is not for token user.",
            )
            raise HTTPException(500, detail=detail)
        elif token_user_grant.level != Level.own:
            detail = dict(
                uuid_user=token_user.uuid,
                uuid_grant=token_user_grant.uuid,
                msg="Grant must be of level `own`.",
                level=token_user_grant.level.name,
            )
            raise HTTPException(500, detail=detail)

        pending_grants = data.data.grants
        for grant in (pending_values := pending_grants.values()):
            if grant.pending_from != PendingFrom.granter:
                raise HTTPException(
                    500,
                    detail="Cannot approve grant pending from `grantee` or `created`.",
                )
            grant.pending = False
            grant.uuid_parent = token_user_grant.uuid
            # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Not assigned when inviting.

        event = self.create_event_grant(data, tuple(pending_values))
        event.detail = "Access granted."
        data.event = event
        return data

    # ------------------------------------------------------------------------ #

    def assignment_document(
        self,
        data: Data[ResolvedAssignmentDocument],
    ) -> Data[ResolvedAssignmentDocument]:
        raise HTTPException(400, detail="Not implemented.")

    def assignment_collection(
        self,
        data: Data[ResolvedAssignmentCollection],
    ) -> Data[ResolvedAssignmentCollection]:
        raise HTTPException(400, detail="Not implemented.")

    # ----------------------------------------------------------------------- #

    def event(self, data: Data[ResolvedEvent]) -> Data[ResolvedEvent]:
        token_user = self.token_user or data.token_user
        (event,) = data.data.events

        event_undo = Event(
            **self.event_common,
            detail="Deletion event reverted.",
            uuid_user=token_user.uuid,  # type: ignore[union-attr]
            kind_obj=KindObject.event,
            uuid_obj=event.uuid,
        )
        event.uuid_undo = event_undo.uuid

        session = self.session
        session.add(event_undo)
        session.add(event)

        for item in event.flattened():
            object_ = item.object_
            if object_ is None or not hasattr(object_, "deleted"):
                continue
            object_.deleted = False

            item.uuid_undo = event_undo.uuid
            session.add(object_)
            session.add(item)

        session.commit()
        session.refresh(event)

        data.event = event_undo
        return data



import httpx
import typer

# --------------------------------------------------------------------------- #
from legere import flags
from legere.handlers import CONSOLE
from legere.requests.base import typerize

from .base import BaseRequests, ContextData, methodize, params


class TokenRequests(BaseRequests):
    typer_commands = dict(
        read="req_read",
        create="req_create",
        register="req_register",
    )

    @classmethod
    def req_register(
        cls,
        _context: typer.Context,
        *,
        email: flags.FlagEmail,
        name: flags.FlagName,
        description: flags.FlagDescription,
        url: flags.FlagUrlOptional = None,
        url_image: flags.FlagUrlImageOptional = None,
        public: flags.FlagPublic = True,
    ) -> httpx.Request:

        context = ContextData.resolve(_context)

        return httpx.Request(
            "POST",
            context.url("/auth/register"),
            headers=context.headers,
            params=params(
                email=email,
                name=name,
                description=description,
                url=url,
                url_image=url_image,
                public=public,
            ),
        )

    @classmethod
    def req_read(
        cls,
        _context: typer.Context,
        # NOTE: now global.
        # token: flags.FlagTokenOptional = None,
    ) -> httpx.Request:
        """Verify current token (as specified by configuration) or, when
        provided, the token provided by `--token`."""

        context = ContextData.resolve(_context)
        token = context.token
        if token is None:
            CONSOLE.print("[red]No token to check.")
            raise typer.Exit(1)

        return httpx.Request(
            "GET",
            context.url("/auth/token"),
            params=params(data=token.get_secret_value()),
            headers=context.headers,
        )

    @classmethod
    def req_create(
        cls,
        _context: typer.Context,
        uuid_user: flags.FlagUUIDUserOptional = None,
        admin: flags.FlagAdmin = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        if uuid_user is not None:
            uuid = uuid_user
        elif context.config.profile is not None:
            uuid = context.config.profile.uuid_user
        else:
            context.console_handler.console.print("[red]Could not determine uuid.")
            raise typer.Exit(1)

        permissions = list()
        if admin:
            permissions.append("tier:admin")

        token_payload = dict(sub=uuid, permissions=permissions)
        return httpx.Request(
            "POST",
            context.url("/auth/token"),
            json=token_payload,
            headers=context.headers,
        )

    typer_check_verbage = False

    read = methodize(req_read, __func__=req_read.__func__)  # type: ignore[attr-defined]
    create = methodize(req_create, __func__=req_create.__func__)  # type: ignore[attr-defined]
    register = methodize(req_register, __func__=req_register.__func__)  # type: ignore[attr-defined]


__all__ = ("TokenRequests",)


if __name__ == "__main__":
    tokens = typerize(TokenRequests)
    tokens()



# =========================================================================== #
import abc
import functools
from http import HTTPMethod
from typing import (
    Any,
    Callable,
    Concatenate,
    Dict,
    Literal,
    ParamSpec,
    Tuple,
    Type,
    TypeVar,
    overload,
)

from fastapi import HTTPException
from sqlalchemy import false, select
from sqlalchemy.orm import Session

# --------------------------------------------------------------------------- #
from captura import __version__
from captura.auth import Token
from captura.controllers.base import (
    BaseController,
    Data,
    DataResolvedAssignment,
    DataResolvedGrant,
    KindData,
    ResolvedAssignmentCollection,
    ResolvedAssignmentDocument,
    ResolvedCollection,
    ResolvedDocument,
    ResolvedEvent,
    ResolvedGrantDocument,
    ResolvedGrantUser,
    ResolvedObjectEvents,
    ResolvedUser,
    T_Data,
)
from captura.err import (
    ErrAccessCollection,
    ErrAccessDocumentCannotRejectOwner,
    ErrAccessUser,
    ErrUpdateGrantPendingFrom,
)
from captura.fields import PendingFrom, Singular
from captura.models import (
    Base,
    Collection,
    Document,
    Event,
    Grant,
    KindObject,
    Level,
    Resolvable,
    ResolvableLevel,
    ResolvableMultiple,
    ResolvableSingular,
    T_Resolvable,
    Tables,
    User,
)
from captura.schemas import EventParams, EventSearchSchema, mwargs

H = HTTPMethod
AccessAssignmentResult = (
    Tuple[Collection, Tuple[Document, ...]] | Tuple[Document, Tuple[Collection, ...]]
)


class Access(BaseController):
    def __call__(
        self,
        kind_data: KindData,
        resolvable: ResolvableSingular,
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: bool = False,
        **kwargs,
    ) -> Data:
        if (method := getattr(self, kind_data.name, None)) is None:
            raise ValueError(f"No such method `{kind_data.name}` of `Access`.")
        return method(
            resolvable,
            resolve_user_token=resolve_user_token,
            exclude_deleted=exclude_deleted,
            return_data=return_data,
            **kwargs,
        )

    # ----------------------------------------------------------------------- #
    # Events

    # NOTE: Order of overloads matters.
    @overload
    def event(
        self,
        resolvable_uuid: Resolvable[Event],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: Literal[True] = True,
    ) -> Data[ResolvedEvent]: ...

    @overload
    def event(
        self,
        resolvable_uuid: ResolvableMultiple[Event],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: Literal[False] = False,
    ) -> Tuple[Event, ...]: ...

    @overload
    def event(
        self,
        resolvable_uuid: ResolvableSingular[Event],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: Literal[False] = False,
    ) -> Event: ...

    def event(
        self,
        resolvable_uuid: Resolvable[Event],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: bool = False,
    ) -> Event | Tuple[Event, ...] | Data[ResolvedEvent]:
        session = self.session
        token_user = self.token_user_or(resolve_user_token)
        events = Event.resolve(session, resolvable_uuid)

        def check_one(event: Event) -> Event:
            token_user.check_can_access_event(event)
            if exclude_deleted:
                event.check_not_deleted()
            return event

        res: Tuple[Event, ...] | Event
        match events:
            case tuple() as items:
                res = tuple(map(check_one, items))
            case Event() as item:
                res = check_one(item)
                if not return_data:
                    return res
            case bad:
                raise ValueError(f"Invalid input `{bad}`.")

        if return_data:
            return mwargs(
                Data[ResolvedEvent],
                data=mwargs(ResolvedEvent, events=events, token_user=token_user),
            )
        return res

    def d_event(
        self,
        resolvable_uuid: Resolvable[Event],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
    ) -> Data[ResolvedEvent]:
        return self.event(
            resolvable_uuid,
            resolve_user_token=resolve_user_token,
            exclude_deleted=exclude_deleted,
            return_data=True,
        )

    @overload
    def object_events(  # type: ignore[overload-overlap]
        self,
        object_resolvable: ResolvableSingular[T_Resolvable],
        object_kind: KindObject,
        param: EventSearchSchema | EventParams | None = None,
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: Literal[True] = True,
    ) -> Data[ResolvedObjectEvents]: ...

    @overload
    def object_events(
        self,
        object_resolvable: ResolvableSingular[T_Resolvable],
        object_kind: KindObject,
        param: EventSearchSchema | EventParams | None = None,
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: Literal[False] = False,
    ) -> Tuple[T_Resolvable, Tuple[Event, ...]]: ...

    def object_events(
        self,
        object_resolvable: ResolvableSingular[T_Resolvable],
        object_kind: KindObject,
        param: EventSearchSchema | EventParams | None = None,
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: bool = False,
    ) -> Data[ResolvedObjectEvents] | Tuple[T_Resolvable, Tuple[Event, ...]]:
        # NOTE: Resolved.
        resolve_user_token = self.token_user_or(resolve_user_token)
        T_Mapped: Type[Base]  # [T_Resolvable]
        T_Mapped = Tables[Singular(object_kind.name).name].value
        object_: T_Resolvable = T_Mapped.resolve(  # type: ignore
            self.session,
            object_resolvable,
        )

        # NOTE: Check access to the object.
        _ = self(
            KindData[object_kind.name],
            object_,
            resolve_user_token=resolve_user_token,
            exclude_deleted=exclude_deleted,
            return_data=False,
        )

        # Find events.
        if param is not None:
            match param:
                case EventParams():
                    exclude = {"root"}
                case EventSearchSchema():
                    exclude = set()

            search = param.model_dump(exclude=exclude)
        else:
            search = dict()

        q_event = Event.q_select_search(**search)
        events = tuple(self.session.execute(q_event).scalars())

        if return_data:
            return mwargs(
                Data[ResolvedObjectEvents],
                data=mwargs(
                    ResolvedObjectEvents,
                    obj=object_,
                    kind_obj=object_kind,
                    events=events,
                ),
            )

        return object_, events

    def d_object_events(
        self,
        object_resolvable: ResolvableSingular[T_Resolvable],
        object_kind: KindObject,
        param: EventSearchSchema | EventParams | None = None,
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
    ) -> Data[ResolvedObjectEvents]:
        return self.object_events(
            object_resolvable,
            object_kind,
            param,
            resolve_user_token=resolve_user_token,
            exclude_deleted=exclude_deleted,
            return_data=True,
        )

    # ----------------------------------------------------------------------- #
    # User

    @overload
    def user(
        self,
        resolve_user: Resolvable[User],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        exclude_public: bool = False,
        return_data: Literal[True] = True,
    ) -> Data[ResolvedUser]: ...

    @overload
    def user(
        self,
        resolve_user: ResolvableMultiple[User],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        exclude_public: bool = False,
        return_data: Literal[False] = False,
    ) -> Tuple[User, ...]: ...

    @overload
    def user(
        self,
        resolve_user: ResolvableSingular[User],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        exclude_public: bool = False,
        return_data: Literal[False] = False,
    ) -> User: ...

    def user(  # type: ignore
        self,
        resolve_user: Resolvable[User],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        exclude_public: bool = False,
        return_data: bool = False,
    ) -> User | Tuple[User, ...] | Data[ResolvedUser]:
        """See if the token user can view another user.

        Resolve the user and verify that is not deleted.
        """

        user_token = self.token_user_or(resolve_user_token)
        users: Tuple[User, ...]
        match res := User.resolve(self.session, resolve_user):
            case tuple() as users:
                users = tuple(
                    self._user(
                        user,
                        user_token,
                        exclude_public=exclude_public,
                        exclude_deleted=exclude_deleted,
                    )
                    for user in users
                )
            case User() as user:
                user = self._user(
                    user,
                    user_token,
                    exclude_deleted=exclude_deleted,
                    exclude_public=exclude_public,
                )
                if not return_data:
                    return user
                users = (user,)
            case _:
                raise HTTPException(405)

        if return_data:
            return Data(
                data=ResolvedUser.model_validate(dict(users=users, kind="user")),  # type: ignore
                token_user=user_token,
                event=None,
            )
        return res

    # NOTE: When `GET` method, if the user is public, return. Otherwise
    #       always check for a token and check the token uuid.
    def _user(
        self,
        user: User,
        user_token: User,
        *,
        exclude_public: bool = False,
        exclude_deleted: bool = True,
        msg_name: (
            Literal[
                "_msg_private",
                "_msg_modify",
                "_msg_only_self",
            ]
            | None
        ) = None,
    ) -> User:
        """
        :param exclude_public: When ``True``, will exclude checking if user is
            public or not in the ``get`` case.
        """
        # if user.admin:
        #     return user
        if exclude_deleted:
            user.check_not_deleted()

        match self.method:
            # case _ if not user.public:
            #     return user
            case H.GET if not exclude_public:
                if (
                    # not exclude_public
                    not user.public
                    and user_token.uuid != user.uuid
                ):
                    raise ErrAccessUser.httpexception(
                        msg_name or "_msg_private",
                        403,
                        uuid_user_token=user_token.uuid,
                        uuid_user=user.uuid,
                    )
                return user
            case H.POST | H.PATCH | H.PUT | H.DELETE | H.GET:
                if user.uuid != user_token.uuid:
                    if self.method == H.GET:
                        msg_name = "_msg_private"

                    raise ErrAccessUser.httpexception(
                        msg_name or "_msg_modify",
                        403,
                        uuid_user=user.uuid,
                        uuid_user_token=user_token.uuid,
                        # msg=msg_access_user,
                    )
                return user
            case _ as bad:
                raise ValueError(f"Cannot yet method `{bad}`.")

    # NOTE: Tried partials. Makes typing hell worse. Easier just to write out.
    #       Tried many other solutions, just do this for now. Should have used
    #       lang with actual overloading.
    def d_user(
        self,
        resolve_user: Resolvable[User],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_public: bool = False,
        exclude_deleted: bool = True,
    ) -> Data[ResolvedUser]:
        return self.user(
            resolve_user,
            resolve_user_token=resolve_user_token,
            return_data=True,
            exclude_deleted=exclude_deleted,
            exclude_public=exclude_public,
        )

    # ----------------------------------------------------------------------- #
    # Collection

    @overload
    def collection(  # type: ignore[overload-overlap]
        self,
        resolve_collection: ResolvableMultiple[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        return_data: Literal[False] = False,
        allow_public: bool = True,
    ) -> Tuple[Collection, ...]: ...

    @overload
    def collection(
        self,
        resolve_collection: Resolvable[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        return_data: Literal[True] = True,
        allow_public: bool = True,
    ) -> Data[ResolvedCollection]: ...

    @overload
    def collection(
        self,
        resolve_collection: ResolvableSingular[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        return_data: Literal[False] = False,
        allow_public: bool = True,
    ) -> Collection: ...

    def collection(
        self,
        resolve_collection: Resolvable[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        return_data: bool = False,
        allow_public: bool = True,
    ) -> Collection | Tuple[Collection, ...] | Data[ResolvedCollection]:
        # NOTE: `exclude_deleted` should only be ``True`` when a force
        #       deletion is occuring.
        def check_one(collection: Collection) -> Collection:
            if exclude_deleted:
                collection = collection.check_not_deleted()

            match self.method:
                case H.GET if allow_public:
                    if not collection.public and collection.id_user != token_user.id:
                        raise ErrAccessCollection.httpexception(
                            "_msg_private",
                            403,
                            uuid_user_token=token_user.uuid,
                            uuid_collection=collection.uuid,
                        )
                    return collection
                case H.GET | H.POST | H.DELETE | H.PUT | H.PATCH:
                    if token_user.id != collection.id_user:
                        raise ErrAccessCollection.httpexception(
                            "_msg_modify",
                            403,
                            uuid_user_token=token_user.uuid,
                            uuid_collection=collection.uuid,
                        )
                    return collection
                case _:
                    msg = f"Cannot handle HTTPMethod `{self.method}`."
                    raise ValueError(msg)

        token_user = self.token_user_or(resolve_user_token)

        collections: Tuple[Collection, ...]
        match res := Collection.resolve(self.session, resolve_collection):
            case Collection():
                _collection = check_one(res)
                if not return_data:
                    return _collection
                collections = (_collection,)
            case tuple():
                collections = tuple(map(check_one, res))
            case _ as bad:
                raise ValueError(
                    "`collections must be a `Collection` or `tuple` of "
                    f"`Collection`s (got `{type(bad)}`)."
                )

        if return_data:
            return mwargs(
                Data[ResolvedCollection],
                data=mwargs(
                    ResolvedCollection,
                    collections=collections,
                    kind="collection",
                ),
                token_user=token_user,
                event=None,
            )
        return res

    def d_collection(
        self,
        resolve_collection: Resolvable[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        allow_public: bool = True,
    ) -> Data[ResolvedCollection]:
        return self.collection(
            resolve_collection,
            exclude_deleted=exclude_deleted,
            resolve_user_token=resolve_user_token,
            allow_public=allow_public,
            return_data=True,
        )

    # ----------------------------------------------------------------------- #
    # Documents

    @overload
    def document(
        self,
        resolve_document: Resolvable[Document],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: Literal[True] = True,
        level: ResolvableLevel | None = None,
        grants: Dict[str, Grant] | None = None,
        grants_index: Literal["uuid_document", "uuid_user"] = "uuid_document",
        pending: bool = False,
        validate: bool = True,
        allow_public: bool = False,
        # pending_from: PendingFrom | None = None,
    ) -> Data[ResolvedDocument]: ...

    @overload
    def document(
        self,
        resolve_document: ResolvableSingular[Document],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: Literal[False] = False,
        level: ResolvableLevel | None = None,
        grants: Dict[str, Grant] | None = None,
        grants_index: Literal["uuid_document", "uuid_user"] = "uuid_document",
        pending: bool = False,
        validate: bool = True,
        allow_public: bool = False,
        # pending_from: PendingFrom | None = None,
    ) -> Document: ...

    @overload
    def document(
        self,
        resolve_document: ResolvableMultiple[Document],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: Literal[False] = False,
        level: ResolvableLevel | None = None,
        grants: Dict[str, Grant] | None = None,
        grants_index: Literal["uuid_document", "uuid_user"] = "uuid_document",
        pending: bool = False,
        validate: bool = True,
        allow_public: bool = False,
        # pending_from: PendingFrom | None = None,
    ) -> Tuple[Document, ...]: ...

    def document(
        self,
        resolve_document: Resolvable[Document],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        return_data: bool = False,
        level: ResolvableLevel | None = None,
        grants: Dict[str, Grant] | None = None,
        grants_index: Literal["uuid_document", "uuid_user"] = "uuid_document",
        pending: bool = False,
        validate: bool = True,
        allow_public: bool = False,
        # pending_from: PendingFrom | None = None,
    ) -> Document | Tuple[Document, ...] | Data[ResolvedDocument]:
        level = level if level is not None else self.level
        token_user = self.token_user_or(resolve_user_token)
        if grants is None:
            grants = dict()

        # NOTE: Exclude deleted is only required for force deletion. Deletedness
        #       exceeds access in terms of priority.
        def check_one(document: Document) -> Document:
            if exclude_deleted:
                document = document.check_not_deleted()
            if allow_public and document.public:
                return document
            else:
                token_user.check_can_access_document(
                    document,
                    level,
                    grants=grants,
                    grants_index=grants_index,
                    pending=pending,
                    exclude_deleted=exclude_deleted,
                    validate=validate,
                )
            return document

        documents: Tuple[Document, ...]
        match Document.resolve(self.session, resolve_document):
            case tuple() as documents:
                documents = tuple(map(check_one, documents))
            case Document() as document:
                _document = check_one(document)
                if not return_data:
                    return _document
                documents = (_document,)
            case _ as bad:
                msg = f"Unexpected input of type `{type(bad)}`."
                raise ValueError(msg)

        if return_data:
            return mwargs(
                Data[ResolvedDocument],
                data=ResolvedDocument.model_validate(
                    dict(
                        documents=documents,
                        kind="document",
                        token_user_grants=grants,
                    )
                ),
                token_user=token_user,
                event=None,
            )
        return documents

    def d_document(
        self,
        resolve_document: Resolvable[Document],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        level: ResolvableLevel | None = None,
        pending: bool = False,
        validate: bool = True,
        allow_public: bool = False,
        # pending_from: PendingFrom | None = None,
    ) -> Data[ResolvedDocument]:
        return self.document(
            resolve_document,
            exclude_deleted=exclude_deleted,
            resolve_user_token=resolve_user_token,
            level=level,
            return_data=True,
            pending=pending,
            validate=validate,
            allow_public=allow_public,
            # pending_from=pending_from,
        )

    # ----------------------------------------------------------------------- #
    # grants

    @overload
    def grant_user(  # type: ignore[overload-overlap]
        self,
        resolve_user: ResolvableSingular[User],
        resolve_documents: ResolvableMultiple[Document] | None,
        *,
        resolve_user_token: ResolvableSingular | None = None,
        exclude_deleted: bool = True,
        level: ResolvableLevel | None = None,
        return_data: Literal[False] = False,
        pending: bool = False,
        validate: bool = False,
        # pending_from: PendingFrom | None = None,
    ) -> Tuple[User, Tuple[Document, ...]]: ...

    @overload
    def grant_user(
        self,
        resolve_user: ResolvableSingular[User],
        resolve_documents: ResolvableMultiple[Document] | None,
        *,
        resolve_user_token: ResolvableSingular | None = None,
        exclude_deleted: bool = True,
        level: ResolvableLevel | None = None,
        return_data: Literal[True] = True,
        pending: bool = False,
        validate: bool = False,
        # pending_from: PendingFrom | None = None,
    ) -> Data[ResolvedGrantUser]: ...

    def grant_user(
        self,
        resolve_user: ResolvableSingular[User],
        resolve_documents: ResolvableMultiple[Document] | None,
        *,
        resolve_user_token: ResolvableSingular | None = None,
        exclude_deleted: bool = True,
        level: ResolvableLevel | None = None,
        return_data: bool = False,
        pending: bool = False,
        validate: bool = False,
        # pending_from: PendingFrom | None = None,
    ) -> Tuple[User, Tuple[Document, ...]] | Data[ResolvedGrantUser]:
        """When inspecting the user, one must authenticate as the user.

        About `PendingFrom`
        -----------------------------------------------------------------------

        Pending from would be nice to have up front (i.e. here) but its value
        only really matters when accepting grants, so the functionality exists
        in the corresponding method (with a similar name) in :class:`Update`.

        More or less, in the update case, the status of the documents would
        have to be checked when :attr:`method` is ``PATCH``.


        About Checking Document Permissions
        -----------------------------------------------------------------------

        Access is generally not required for the documents in
        ``resolve_documents``. However, it would still be useful to populate
        ``grants`` using ``self.document``. For this reason the parameter
        ``validate`` is added.
        """

        level = Level.resolve(level) if level is not None else self.level
        user_token = self.token_user_or(resolve_user_token)

        user = self._user(
            User.resolve(self.session, resolve_user),
            user_token,
            exclude_deleted=exclude_deleted,
            exclude_public=True,
            msg_name="_msg_only_self",
        )
        document_kwargs: Dict[str, Any] = dict(
            exclude_deleted=exclude_deleted,
            level=level,
            pending=pending,
        )

        uuid_documents = (
            Document.resolve_uuid(self.session, resolve_documents)
            if resolve_documents is not None
            else None
        )

        # NOTE: In the case of post, a document just has to not be deleted.
        #       It is easier (for now) to just write out the query here.
        q: Any
        if self.method == H.POST:
            if uuid_documents is None:
                raise ValueError("`uuid_documents` is `None`.")
            q = select(Document).where(Document.uuid.in_(uuid_documents))
            q = q.where(Document.deleted == false())
            resolve_documents = tuple(self.session.scalars(q))
        else:
            q = user.q_select_documents(
                uuid_documents,
                **document_kwargs,
                exclude_pending=not pending,
            )
            resolve_documents = tuple(self.session.execute(q).scalars())

        token_user_grants: Dict[str, Grant] = dict()
        documents = self.document(
            resolve_documents,
            **document_kwargs,
            resolve_user_token=user_token,
            grants=token_user_grants,
            grants_index="uuid_document",
            validate=validate,
            # pending_from=pending_from,
        )

        # NOTE: If the token user is not the user that the data is for then
        #       it is necessary to find a different set of grants.
        grants_user: Dict[str, Grant]
        if user_token.uuid != user.uuid:
            _ = self.document(
                documents,
                **document_kwargs,
                resolve_user_token=user,
                grants=(grants_user := dict()),
                grants_index="uuid_document",
            )
        else:
            grants_user = token_user_grants

        if self.method == H.PATCH and len(
            bad := {
                uuid_document
                for uuid_document, grant in grants_user.items()
                if grant.pending_from != PendingFrom.grantee
            }
        ):
            # NOTE: Ensure all documents have the currect pending from value
            #       for approval
            raise ErrUpdateGrantPendingFrom.httpexception(
                "_msg_grantee",
                403,
                uuid_obj=bad,
                kind_obj=KindObject.document,
            )

        if return_data:
            return mwargs(
                Data[ResolvedGrantUser],
                data=mwargs(
                    ResolvedGrantUser,
                    documents=documents,
                    user=user,
                    grants=grants_user,
                    kind="grant_user",
                    token_user=self.token_user,
                    token_user_grants=token_user_grants,
                ),
                token_user=user_token,
                event=None,
            )
        return user, documents

    def d_grant_user(
        self,
        resolve_user: ResolvableSingular[User],
        resolve_documents: ResolvableMultiple[Document] | None,
        *,
        resolve_user_token: ResolvableSingular | None = None,
        exclude_deleted: bool = True,
        # exclude_pending: bool = True,
        level: ResolvableLevel | None = None,
        pending: bool = False,
        validate: bool = False,
    ) -> Data[ResolvedGrantUser]:
        return self.grant_user(
            resolve_user,
            resolve_documents,
            exclude_deleted=exclude_deleted,
            resolve_user_token=resolve_user_token,
            level=level,
            return_data=True,
            pending=pending,
            validate=validate,
        )

    @overload
    def grant_document(
        self,
        resolve_document: ResolvableSingular[Document],
        resolve_users: ResolvableMultiple[User] | None,
        *,
        resolve_user_token: ResolvableSingular | None = None,
        exclude_deleted: bool = True,
        exclude_pending: bool = False,
        level: ResolvableLevel | None = None,
        return_data: Literal[False] = False,
        pending: bool = False,
    ) -> Tuple[Document, Tuple[User, ...]]: ...

    @overload
    def grant_document(
        self,
        resolve_document: ResolvableSingular[Document],
        resolve_users: ResolvableMultiple[User] | None,
        *,
        resolve_user_token: ResolvableSingular | None = None,
        exclude_deleted: bool = True,
        exclude_pending: bool = False,
        level: ResolvableLevel | None = None,
        return_data: Literal[True] = True,
        pending: bool = False,
    ) -> Data[ResolvedGrantDocument]: ...

    def grant_document(
        self,
        resolve_document: ResolvableSingular[Document],
        resolve_users: ResolvableMultiple[User] | None,
        *,
        resolve_user_token: ResolvableSingular | None = None,
        exclude_deleted: bool = True,
        exclude_pending: bool = False,
        level: ResolvableLevel | None = None,
        return_data: bool = False,
        pending: bool = False,
    ) -> Tuple[Document, Tuple[User, ...]] | Data[ResolvedGrantDocument]:
        """For document owners only."""

        level = Level.resolve(level) if level is not None else self.level
        user_token = self.token_user_or(resolve_user_token)

        token_user_grant: Dict[str, Grant] = dict()
        document = self.document(
            resolve_document,
            exclude_deleted=exclude_deleted,
            resolve_user_token=user_token,
            return_data=False,
            level=Level.own,
            pending=pending,
            grants=token_user_grant,
            grants_index="uuid_user",
        )

        # NOTE: Permissions of users do not matter for CRUD of grants because
        #       they're always url params that are filtered out. Notice:
        #
        #       1. Reading the grants for document requires no knowledge
        #          about the users filtered by.
        #       2. Creating pending grants for a document requires nothing
        #          about the various users to be known with regard to the
        #          source document.
        #       3. Accepting requests for access does not require any grant
        #          knowledge bout the target users on the source document.
        #
        # NOTE: To get grants just use `ResolvedGrant*.grants`.
        users: Tuple[User, ...]
        if resolve_users is None:
            q = document.q_select_users(
                level=Level.own,
                exclude_deleted=exclude_deleted,
                pending=pending,
                exclude_pending=False,
            )
            users = tuple(self.session.execute(q).scalars())
        else:
            users = User.resolve(self.session, resolve_users)

        uuid_users = User.resolve_uuid(self.session, users)
        match self.method:
            # NOTE: Ensure all documents have the currect pending from value
            #       for approval
            case H.GET | H.POST | H.PUT | H.PATCH:
                ...
            case H.DELETE:
                session = self.session
                q_owners = document.q_select_users(
                    level=Level.own,
                    exclude_deleted=exclude_deleted,
                    # pending=pending,
                    exclude_pending=True,  # Absolutely necessary, do not rm
                )

                uuid_owners: set[str]
                uuid_owners = set(item.uuid for item in session.scalars(q_owners))
                if user_token.uuid in uuid_owners:
                    uuid_owners.remove(user_token.uuid)

                if len(uuid_bad := uuid_owners & uuid_users):
                    raise ErrAccessDocumentCannotRejectOwner.httpexception(
                        "_msg_cannot_reject_owner",
                        403,
                        uuid_user_revoker=user_token.uuid,
                        uuid_user_revokees=uuid_bad,
                        uuid_document=Document.resolve_uuid(session, resolve_document),
                    )
            case _:
                raise HTTPException(405)

        q_user_grants = document.q_select_grants(
            uuid_users,
            level=level,
            exclude_pending=exclude_pending,
            pending=pending,
            exclude_deleted=exclude_deleted,
        )
        user_grant: Dict[str, Grant] = {
            grant.uuid_user: grant for grant in self.session.scalars(q_user_grants)
        }

        # NOTE: Check that the grants in question is.
        if self.method == H.PATCH and len(
            bad := {
                uuid_user
                for uuid_user, grant in user_grant.items()
                if grant.pending_from != PendingFrom.granter
            }
        ):
            raise ErrUpdateGrantPendingFrom.httpexception(
                "_msg_granter",
                403,
                uuid_obj=bad,
                kind_obj=KindObject.user,
            )

        if return_data:
            return mwargs(
                Data,
                data=mwargs(
                    ResolvedGrantDocument,
                    document=document,
                    grants=user_grant,
                    users=users,
                    kind="grant_document",
                    token_user_grants=token_user_grant,
                ),
                token_user=user_token,
            )
        return document, users

    def d_grant_document(
        self,
        resolve_document: ResolvableSingular[Document],
        resolve_users: ResolvableMultiple[User] | None,
        *,
        resolve_user_token: ResolvableSingular | None = None,
        exclude_deleted: bool = True,
        exclude_pending: bool = True,
        level: ResolvableLevel | None = None,
        pending: bool = False,
    ) -> Data[ResolvedGrantDocument]:
        return self.grant_document(
            resolve_document,
            resolve_users,
            exclude_deleted=exclude_deleted,
            resolve_user_token=resolve_user_token,
            level=level,
            return_data=True,
            pending=pending,
            exclude_pending=exclude_pending,
        )

    # ----------------------------------------------------------------------- #
    # Assignments

    @overload
    def assignment_collection(  # type: ignore[overload-overlap]
        self,
        resolve_collection: ResolvableSingular[Collection],
        resolve_documents: ResolvableMultiple[Document],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        level: ResolvableLevel | None = None,
        return_data: Literal[False] = False,
        allow_public: bool = False,
        allow_public_collection: bool = False,
        validate_documents: bool = True,
    ) -> Tuple[Collection, Tuple[Document, ...]]: ...

    @overload
    def assignment_collection(
        self,
        resolve_collection: ResolvableSingular[Collection],
        resolve_documents: ResolvableMultiple[Document],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        level: ResolvableLevel | None = None,
        return_data: Literal[True] = True,
        allow_public: bool = False,
        allow_public_collection: bool = False,
        validate_documents: bool = True,
    ) -> Data[ResolvedAssignmentCollection]: ...

    def assignment_collection(
        self,
        resolve_collection: ResolvableSingular[Collection],
        resolve_documents: ResolvableMultiple[Document],
        *,
        resolve_user_token: ResolvableSingular[User] | None = None,
        exclude_deleted: bool = True,
        level: ResolvableLevel | None = None,
        return_data: bool = False,
        allow_public: bool = False,
        allow_public_collection: bool = False,
        validate_documents: bool = True,
    ) -> Tuple[Collection, Tuple[Document, ...]] | Data[ResolvedAssignmentCollection]:
        # NOTE: Keep `token_user` here so that the user is checked.
        token_user = self.token_user_or(resolve_user_token)
        collection = self.collection(
            resolve_collection,
            exclude_deleted=exclude_deleted,
            resolve_user_token=token_user,
            allow_public=allow_public_collection,
            return_data=False,
        )
        documents = self.document(
            resolve_documents,
            level=level or self.level,
            exclude_deleted=exclude_deleted,
            resolve_user_token=token_user,
            allow_public=allow_public,
            validate=validate_documents,
            return_data=False,
        )

        uuid_documents = Document.resolve_uuid(self.session, resolve_documents)
        q_assignments = collection.q_select_assignment(uuid_documents)
        assignments = {
            assignment.uuid_document: assignment
            for assignment in self.session.execute(q_assignments).scalars()
        }

        if return_data:
            return mwargs(
                Data[ResolvedAssignmentCollection],
                data=mwargs(
                    ResolvedAssignmentCollection,
                    kind="assignment_collection",
                    collection=collection,
                    assignments=assignments,
                    documents=documents,
                    uuid_documents=uuid_documents,
                ),
                token_user=token_user,
                event=None,
            )
        return collection, documents

    def d_assignment_collection(
        self,
        resolve_collection: ResolvableSingular[Collection],
        resolve_documents: ResolvableMultiple[Document],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        level: ResolvableLevel | None = None,
        validate_documents: bool = True,
        allow_public: bool = False,
        allow_public_collection: bool = False,
    ) -> Data[ResolvedAssignmentCollection]:
        return self.assignment_collection(
            resolve_collection,
            resolve_documents,
            exclude_deleted=exclude_deleted,
            resolve_user_token=resolve_user_token,
            level=level,
            return_data=True,
            validate_documents=validate_documents,
            allow_public=allow_public,
            allow_public_collection=allow_public_collection,
        )

    @overload
    def assignment_document(  # type: ignore[overload-overlap]
        self,
        resolve_document: ResolvableSingular[Document],
        resolve_collections: ResolvableMultiple[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        level: ResolvableLevel | None = None,
        return_data: Literal[False] = False,
        validate_collections: bool = True,
        validate_document: bool = True,
        allow_public: bool = False,
        all_collections: bool = False,
    ) -> Tuple[Document, Tuple[Collection, ...]]: ...

    @overload
    def assignment_document(
        self,
        resolve_document: ResolvableSingular[Document],
        resolve_collections: ResolvableMultiple[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        level: ResolvableLevel | None = None,
        return_data: Literal[True] = True,
        validate_collections: bool = True,
        validate_document: bool = True,
        allow_public: bool = False,
        all_collections: bool = False,
    ) -> Data[ResolvedAssignmentDocument]: ...

    def assignment_document(
        self,
        resolve_document: ResolvableSingular[Document],
        resolve_collections: ResolvableMultiple[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        level: ResolvableLevel | None = None,
        return_data: bool = False,
        validate_collections: bool = True,
        validate_document: bool = True,
        allow_public: bool = False,
        all_collections: bool = False,
    ) -> Tuple[Document, Tuple[Collection, ...]] | Data[ResolvedAssignmentDocument]:
        token_user = self.token_user_or(resolve_user_token)
        document = self.document(
            resolve_document,
            level=level or self.level,
            exclude_deleted=exclude_deleted,
            resolve_user_token=token_user,
            allow_public=allow_public,
            validate=validate_document,
            return_data=False,
        )
        if validate_collections:
            collections = self.collection(
                resolve_collections,
                exclude_deleted=exclude_deleted,
                resolve_user_token=token_user,
            )
        else:
            collections = Collection.resolve(self.session, resolve_collections)

        if all_collections:
            collections = tuple(
                collection
                for collection in collections
                if collection.public or collection.uuid_user == token_user.uuid
            )

        uuid_collections = Collection.resolve_uuid(self.session, resolve_collections)
        q_assignments = document.q_select_assignment(
            uuid_collections, exclude_deleted=exclude_deleted
        )
        assignments = {
            assignment.uuid_collection: assignment
            for assignment in self.session.execute(q_assignments).scalars()
        }

        if return_data:
            return mwargs(
                Data,
                data=mwargs(
                    ResolvedAssignmentDocument,
                    kind="assignment_document",
                    document=document,
                    collections=collections,
                    assignments=assignments,
                ),
                token_user=token_user,
                event=None,
            )

        return document, collections

    def d_assignment_document(
        self,
        resolve_document: ResolvableSingular[Document],
        resolve_collections: ResolvableMultiple[Collection],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        level: ResolvableLevel | None = None,
        validate_collections: bool = True,
        allow_public: bool = False,
        validate_document: bool = True,
        all_collections: bool = False,
    ) -> Data[ResolvedAssignmentDocument]:
        return self.assignment_document(
            resolve_document,
            resolve_collections,
            exclude_deleted=exclude_deleted,
            resolve_user_token=resolve_user_token,
            level=level,
            return_data=True,
            validate_collections=validate_collections,
            allow_public=allow_public,
            validate_document=validate_document,
            all_collections=all_collections,
        )

    def assignment(
        self,
        source: Document | Collection,
        resolve_target: ResolvableMultiple[Collection] | ResolvableMultiple[Document],
        *,
        exclude_deleted: bool = True,
        resolve_user_token: ResolvableSingular[User] | None = None,
        level: ResolvableLevel | None = None,
    ) -> Data[ResolvedAssignmentDocument] | Data[ResolvedAssignmentCollection]:
        match [source, resolve_target]:
            case [Document() as document, set() | tuple() as targets]:
                return self.assignment_document(
                    document,
                    targets,
                    exclude_deleted=exclude_deleted,
                    resolve_user_token=resolve_user_token,
                    level=level,
                    return_data=True,
                )
            case [Collection() as collection, set() | tuple() as targets]:
                return self.assignment_collection(
                    collection,
                    targets,
                    exclude_deleted=exclude_deleted,
                    resolve_user_token=resolve_user_token,
                    level=level,
                    return_data=True,
                )
            case _ as bad:
                raise ValueError(f"Unexpected source `{bad}`.")


# =========================================================================== #
# Decorators and ABC.
#
# NOTE: While it is possible to write out callabels with aliases, the LSP
#       feedback is garbage (event though they have correct signatures).
T_WithAccess = TypeVar("T_WithAccess", bound="WithAccess")
P_WithAccess = ParamSpec("P_WithAccess")


def with_access(
    fn_access: Callable[
        Concatenate[Access, P_WithAccess],
        Data,
    ],
) -> Callable[
    [Callable[[T_WithAccess, Data], Data]],
    Callable[Concatenate[T_WithAccess, P_WithAccess], Data],
]:
    """Chain together :param:`method` with the appropriate access method.

    The intended result is to avoid having to write this so often:

    .. code:: python

        def doit(
            resolvable_collection: ResolvableSingular[Collection],
            resolvable_documents: ResolvableMultiple[Document],
        ):
            access = Access(session, token, method)
            collection, documents = access.assignment(
                resolvable_collection,
                resolvable_documents
            )
            delete = access.then(Delete)
            event = delete(collection, documents)

    when the access controller was not included in deleted. This MUST be
    used in combination with `WithAccess` so that the access controller is
    available from `self`.

    :param fn: Callable to decorate.
    """

    def __call__(
        fn: Callable[[T_WithAccess, Data], Data]
    ) -> Callable[Concatenate[T_WithAccess, P_WithAccess], Data]:
        @functools.wraps(fn_access)
        def wrapper(
            self,
            *args: P_WithAccess.args,
            **kwargs: P_WithAccess.kwargs,
        ) -> Data:
            # NOTE: `data` should be resolved resolvables. This is why the
            #       signatures must match, and further, the output and input
            #       of all controller methods will resolavble data and resolved
            #       outputs.
            data = fn_access(self.access, *args, **kwargs)
            return fn(self, data)

        return wrapper

    return __call__


# NOTE: Decoration -> Simplified signatures. Other methods will be added with
#       prefixes, e.g. ``a_assignment_document``.
class WithAccess(BaseController, abc.ABC):
    # ----------------------------------------------------------------------- #
    access: Access
    api_origin: str
    force: bool = True

    @property
    def event_common(self) -> Dict[str, Any]:
        # TODO: Should use super.
        data = dict(
            api_origin=self.api_origin,
            api_version=__version__,
        )
        # NOTE: Depending on token user might be a mistake.
        if self._token is not None:
            data["uuid_user"] = self.token_user.uuid
        return data

    # NOTE: Why `Resolved` kind names must correspond to their functions. At
    #       this point I am regretting not using a compiled language. This is
    #       here mostly for the reason that in some instances data might be one
    #       of many kinds, therefore it makes sense to be able to excute
    #       functions this way.
    def __call__(self, data: Data[T_Data]) -> Data[T_Data]:
        method_name: str = data.kind
        if (method := getattr(self, method_name, None)) is None:
            raise ValueError(f"`Access` has no method `{method_name}`.")
        return method(data)

    def __init__(
        self,
        session: Session,
        token: Token | Dict[str, Any] | None,
        method: HTTPMethod | str,
        *,
        api_origin: str,
        force: bool = False,
        access: Access | None = None,
    ):
        super().__init__(session, token, method)
        self.force = force
        self.api_origin = api_origin
        self.access = access if access is not None else self.then(Access)

    # ----------------------------------------------------------------------- #
    # Assignments

    @abc.abstractmethod
    def assignment_collection(
        self, data: Data[ResolvedAssignmentCollection]
    ) -> Data[ResolvedAssignmentCollection]: ...

    @abc.abstractmethod
    def assignment_document(
        self, data: Data[ResolvedAssignmentDocument]
    ) -> Data[ResolvedAssignmentDocument]: ...

    # NOTE: Should never overwrite! Messes up overloads!
    @overload
    def assignment(
        self, data: Data[ResolvedAssignmentCollection]
    ) -> Data[ResolvedAssignmentCollection]: ...

    @overload
    def assignment(
        self, data: Data[ResolvedAssignmentDocument]
    ) -> Data[ResolvedAssignmentDocument]: ...

    def assignment(self, data: DataResolvedAssignment) -> DataResolvedAssignment:
        meth = (
            self.assignment_collection
            if data.kind == "assignment_collection"
            else self.assignment_document
        )
        return meth(data)  # type: ignore

    # ----------------------------------------------------------------------- #
    # Grants

    @abc.abstractmethod
    def grant_user(
        self,
        data: Data[ResolvedGrantUser],
    ) -> Data[ResolvedGrantUser]: ...

    @abc.abstractmethod
    def grant_document(
        self,
        data: Data[ResolvedGrantDocument],
    ) -> Data[ResolvedGrantDocument]: ...

    @overload
    def grant(self, data: Data[ResolvedGrantUser]) -> Data[ResolvedGrantUser]: ...

    @overload
    def grant(
        self, data: Data[ResolvedGrantDocument]
    ) -> Data[ResolvedGrantDocument]: ...

    def grant(self, data: DataResolvedGrant) -> DataResolvedGrant:
        meth = self.grant_user if data.kind == "grant_user" else self.grant_document
        return meth(data)  # type: ignore

    # ----------------------------------------------------------------------- #
    # Everything else

    @abc.abstractmethod
    def user(self, data: Data[ResolvedUser]) -> Data[ResolvedUser]: ...

    @abc.abstractmethod
    def document(
        self,
        data: Data[ResolvedDocument],
    ) -> Data[ResolvedDocument]: ...

    @abc.abstractmethod
    def collection(
        self,
        data: Data[ResolvedCollection],
    ) -> Data[ResolvedCollection]: ...



# =========================================================================== #
import enum
from typing import Annotated, Any, Dict, Literal, Self

import yaml
from pydantic import (
    BaseModel,
    Field,
    SecretStr,
    computed_field,
    model_validator,
)
from yaml_settings_pydantic import (
    BaseYamlSettings,
    YamlFileConfigDict,
    YamlSettingsConfigDict,
)

# --------------------------------------------------------------------------- #
from captura import util

# NOTE: Previously this was part of the console handler. However the same
#       data is necessary in a number of places and is therefore factored out.


class Output(str, enum.Enum):
    raw = "raw"
    json = "json"
    yaml = "yaml"
    table = "table"


class OutputConfig(BaseModel):

    decorate: Annotated[bool, Field(default=True)]
    output: Annotated[Output, Field(default=Output.yaml)]
    output_fallback: Annotated[
        Literal[Output.yaml, Output.json],
        Field(
            default=Output.yaml,
            description="Because some data cannot be nicely rendered as a table.",
        ),
    ]
    rich_theme: str = "fruity"

    # output_exclude: Annotated[Set[str], Field(default_factory=set)]
    # table_columns: Annotated[FlagColumns, Field(default_factory=list)]
    # table_column_configs: Annotated[
    #     Dict[str, Dict[str, Any]], Field(default_factory=dict)
    # ]


class ProfileConfig(BaseModel):
    # name: Annotated[str, Field()]
    uuid_user: Annotated[str, Field()]
    token: Annotated[SecretStr | None, Field(default=None)]


class HostConfig(BaseModel):
    host: Annotated[str, Field(default="http://localhost:8080")]
    remote: Annotated[bool, Field(default=False)]


class UseConfig(BaseModel):
    host: Annotated[str, Field(default="default")]
    profile: Annotated[str, Field(default="default")]


class Config(BaseYamlSettings):
    # NOTE
    model_config = YamlSettingsConfigDict(  # type: ignore
        yaml_files={
            util.PATH_CONFIG_CLIENT: YamlFileConfigDict(
                envvar=util.prefix_env("CONFIG_CLIENT"),
            )
        },
        extra="allow",
    )

    output: Annotated[OutputConfig, Field()]
    profiles: Annotated[Dict[str, ProfileConfig], Field()]
    hosts: Annotated[Dict[str, HostConfig], Field()]
    use: Annotated[UseConfig, Field()]

    # NOTE: See https://github.com/acederberg/pydantic-settings-yaml/issues/22.
    def dump(self, config_path: str) -> None:
        data = self.model_dump_config()
        # print()
        # print("dump", data["output"])
        # print("dump", config_path)
        # print()
        with open(config_path, "w") as file:
            yaml.dump(data, file)

    @model_validator(mode="after")
    def check_use_valid(self):
        fmt = "Invalid %s config `%s`, should be any of `%s`."
        if (uu := self.use.host) not in (vv := self.hosts):
            raise ValueError(fmt % ("host", uu, vv))

        if (uu := self.use.profile) not in (vv := self.profiles):
            raise ValueError(fmt % ("profile", uu, vv))

        return self

    # NOTE: See https://github.com/acederberg/pydantic-settings-yaml/issues/22.
    @classmethod
    def load(cls, config_path: str) -> Self:
        with open(config_path, "r") as file:
            config = cls.model_validate(yaml.safe_load(file))
        return config

    @computed_field  # type: ignore[prop-decorator]
    @property
    def host(self) -> HostConfig:
        if (hh := self.hosts.get(self.use.host)) is None:
            raise ValueError()

        return hh

    @computed_field  # type: ignore[prop-decorator]
    @property
    def profile(self) -> ProfileConfig:
        if (pp := self.profiles.get(self.use.profile)) is None:
            raise ValueError()

        return pp

    @computed_field  # type: ignore[prop-decorator]
    @property
    def token(self) -> SecretStr | None:
        if (pp := self.profile) is None:
            return None

        return pp.token

    def model_dump_minimal(self) -> Dict[str, Any]:
        "Create minimal output for display."

        data = self.model_dump(
            mode="json",
            include={"profile", "host", "output"},
        )
        return data

    def model_dump_config(self) -> Dict[str, Any]:
        "Create file output as a python dictionary."

        data = self.model_dump(
            mode="json",
            include={"profiles", "hosts", "output", "use"},
        )
        for profile_name, profile in data["profiles"].items():
            token_secret = self.profiles[profile_name].token
            token = (
                token_secret.get_secret_value() if token_secret is not None else None
            )
            profile["token"] = token

        return data



# =========================================================================== #
import asyncio
import functools
from collections.abc import Awaitable
from typing import (
    Annotated,
    Any,
    Callable,
    ClassVar,
    Concatenate,
    Dict,
    Generator,
    ParamSpec,
    Self,
    Tuple,
    Type,
    TypeVar,
)
from urllib import parse

import httpx
import typer
import yaml
from click.core import Context as ClickContext
from fastapi.openapi.models import OpenAPI, PathItem
from pydantic import BaseModel, ConfigDict, SecretStr, computed_field
from rich.table import Table
from typer.cli import Command
from typing_extensions import Doc

# --------------------------------------------------------------------------- #
from captura.fields import Singular
from captura.schemas import mwargs
from legere import flags
from legere.config import Config
from legere.flags import Verbage
from legere.handlers import (
    CONSOLE,
    AssertionHandler,
    ConsoleHandler,
    RequestHandlerData,
    render_request,
)

# Helpers for decorators.

url_chunks_accepted = {"users", "documents", "collections", "assignments"}
url_chunks_accepted |= {"grants", "extensions", "demos"}


def openapi_match(req: httpx.Request) -> Generator[str, None, None]:
    chunks = req.url.path.split("/")[1:]

    last_chunk = None
    for chunk in chunks:
        if chunk in url_chunks_accepted:
            yield chunk
            last_chunk = chunk
        else:
            if last_chunk is None:
                raise ValueError(
                    f"Last chunk not yet specified (the current chunk is `{chunk}`)."
                )
            elif last_chunk == "uuid":
                raise ValueError(
                    "Cannot determine uuid from last_chunk since it was a "
                    f"uuid (the current chunk is `{chunk}`)."
                )

            if last_chunk == "demos":
                yield "invitation_uuid"
            else:
                yield f"{{uuid_{Singular[last_chunk].value}}}"
                last_chunk = "uuid"


def openapi_find(openapi: OpenAPI, req: httpx.Request) -> PathItem:
    url = req.url
    path = str(url.path)
    assert path is not None
    assert openapi.paths is not None

    fpath = "/" + "/".join(openapi_match(req))
    httpmethod = req.method
    try:
        for_path = openapi.paths[fpath]
        for_method = for_path.get(httpmethod.lower)  # type: ignore
        return for_method
    except KeyError as err:
        CONSOLE.print(
            f"[red]Could not find data in OpenAPI schema for "
            f"`{httpmethod} {fpath}`."
        )
        raise typer.Exit(1) from err


# =========================================================================== #
# Data for `ClickContext`.


class ContextData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    config: Config
    console_handler: ConsoleHandler

    # NOTE: These flags will change how wrapped functions work.
    openapi: flags.FlagOpenApi = False
    show_request: flags.FlagShowRequest = False
    auth_exclude: flags.FlagNoAuthorization = False
    token_from_global: SecretStr | None = None

    def openapijson(self, client: httpx.Client) -> OpenAPI:
        res = client.send(self.req_openapijson())
        return OpenAPI.model_validate(res.json())

    def req_openapijson(self) -> httpx.Request:
        res = httpx.Request(
            "GET",
            url=self.url("/openapi.json"),
        )
        return res

    @computed_field  # type: ignore[prop-decorator]
    @property
    def token(self) -> SecretStr | None:
        return self.token_from_global or self.config.token

    @computed_field  # type: ignore[prop-decorator]
    @property
    def headers(self) -> Dict[str, str]:
        h = dict(content_type="application/json")
        if not self.auth_exclude and (token := self.token) is not None:
            h.update(authorization=f"bearer {token.get_secret_value()}")
        return h

    @classmethod
    def resolve(cls, ctx: typer.Context | Self) -> Self:
        match ctx:
            case ClickContext():
                if ctx.obj is None:
                    raise ValueError("Context missing `object`.")
                return ctx.obj
            case cls():  # type: ignore
                return ctx
            case bad:
                raise ValueError(f"Cannot resolve context from `{bad}`.")

    def get_config(self, config_path: str | None = None) -> Config:
        """When provided an additional ``config_path``, use said config
        instead. This is included since some commands overwrite the global
        ``--config`` flag.
        """
        if config_path is not None:
            config = Config.load(config_path)
        else:
            config = self.config

        return config

    def url(self, rest: str) -> str:
        if not self.config.host:
            raise ValueError("Host missing.")

        return parse.urljoin(self.config.host.host, rest)

    @classmethod
    def for_typer(
        cls,
        context: typer.Context,
        # Config general,
        path_config: flags.FlagConfig = None,
        profile: flags.FlagProfile = None,
        host: flags.FlagHost = None,
        token: flags.FlagTokenOptional = None,
        # For output
        output: flags.FlagOutput | None = None,
        decorate: flags.FlagDecorate = None,
        # For rest.
        openapi: flags.FlagOpenApi = False,
        show_request: flags.FlagShowRequest = False,
        auth_exclude: flags.FlagNoAuthorization = False,
    ):
        if path_config is None:
            config = mwargs(Config)
        else:
            with open(path_config, "r") as file:
                config = Config.model_validate(yaml.safe_load(file))

        assert config is not None

        if host is not None:
            config.use.host = host
        if profile is not None:
            config.use.profile = profile

        if output is not None:
            # TODO: Figure out why this type hint problem occurs.
            config.output.output = output  # type: ignore
        if decorate is not None:
            config.output.decorate = decorate

        console_handler = ConsoleHandler(config=config)
        context.obj = mwargs(
            ContextData,
            config=config,
            console_handler=console_handler,
            show_request=show_request,
            auth_exclude=auth_exclude,
            token_from_global=token,
        )
        context.obj.openapi = openapi


# =========================================================================== #
# Decorators


P_Wrapped = ParamSpec("P_Wrapped")

MkRequest = Callable[Concatenate[typer.Context | ContextData, P_Wrapped], httpx.Request]
MkRequestTyperized = Callable[Concatenate[typer.Context, P_Wrapped], httpx.Response]


T_Wrapped = TypeVar("T_Wrapped", bound="BaseRequests")
MkRequestCls = (
    Callable[
        Concatenate[Type[T_Wrapped], typer.Context | ContextData, P_Wrapped],
        httpx.Request,
    ]
    | Callable[Concatenate[Type[T_Wrapped], typer.Context, P_Wrapped], httpx.Request]
)
MkRequestInstance = Callable[
    Concatenate[T_Wrapped, P_Wrapped], Awaitable[httpx.Response]
]


# --------------------------------------------------------------------------- #
# Test client stuff.


# NOTE: This decorator is an abomination. The problem is that we want to
#       convert a classmethod into an instance method.
# NOTE: The type hint of fn cannot include `ContextData | typer.Context` bc
#       typer. ``__func__`` is included because classmethods are processed
#       after the end of the class definition.
def methodize(
    fn: MkRequestCls[T_Wrapped, P_Wrapped],
    __func__: Callable | None = None,
) -> MkRequestInstance[T_Wrapped, P_Wrapped]:
    if __func__ is not None:
        fn = __func__  # type: ignore

    @functools.wraps(fn)
    async def wrapper(
        self: T_Wrapped,
        *args: P_Wrapped.args,
        **kwargs: P_Wrapped.kwargs,
    ) -> httpx.Response:
        req = fn(self.__class__, self.context, *args, **kwargs)  # type: ignore
        res = await self.client.send(req)

        if self.handler_methodize:
            self.handler(res, **self.create_handler_args())
        return res

    return wrapper


# --------------------------------------------------------------------------- #
# Typer Stuff


def typerize_fn_httx(
    fn: MkRequest[P_Wrapped],
) -> MkRequestTyperized[P_Wrapped]:
    @functools.wraps(fn)
    def wrapper(
        _context: typer.Context,
        *args: P_Wrapped.args,
        **kwargs: P_Wrapped.kwargs,
    ) -> httpx.Response:
        context = ContextData.resolve(_context)
        with httpx.Client() as client:
            request: httpx.Request = fn(_context, *args, **kwargs)
            if context.openapi:
                res = openapi_find(context.openapijson(client), request)
                status = context.console_handler.handle(None, data=res)
                raise typer.Exit(status)

            if context.show_request:
                context.console_handler.console.print(
                    render_request(context.config.output, request)
                )
                raise typer.Exit(1)

            response = client.send(request)

        context.console_handler.handle(response)
        return response

    return wrapper


def typerize(
    cls: Type["BaseTyperizable"],
    *,
    exclude_callback: bool = False,
    callback=None,
    typerize_fn=typerize_fn_httx,
) -> typer.Typer:

    callback = ContextData.for_typer if callback is None else callback
    callback = callback if not exclude_callback else None

    tt = typer.Typer(help=cls.typer_help or cls.__doc__)
    if not exclude_callback:
        assert callback is not None
        tt.callback()(callback)

    for command_name, command_name_fn in cls.typer_commands.items():
        if cls.typer_check_verbage:
            if command_name not in Verbage._member_map_:
                raise ValueError(f"Illegal verbage `{command_name}`.")

        if (command_fn := getattr(cls, command_name_fn, None)) is None:
            raise ValueError(f"No function `{command_name_fn}`.")

        if cls.typer_decorate:
            command_fn = typerize_fn(command_fn)
        tt.command(command_name)(command_fn)

    for group_name, group_cls in cls.typer_children.items():
        tt.add_typer(
            typerize(group_cls, callback=callback, exclude_callback=True),
            name=group_name,
        )

    return tt


# =========================================================================== #
# BaseRequests


class BaseTyperizable:
    typer_help: ClassVar[str | None] = None
    typer_decorate: ClassVar[bool] = True
    typer_check_verbage: ClassVar[bool] = True
    typer_commands: ClassVar[Dict[str, str]] = dict()
    typer_children: ClassVar[Dict[str, Type["BaseTyperizable"]]] = dict()

    def render_table(self) -> Table | None: ...


class BaseRequests(BaseTyperizable):

    context: ContextData
    client: httpx.AsyncClient
    handler: AssertionHandler
    handler_methodize: Annotated[
        bool,
        Doc(
            "When ``True``, methodized functions will use this callback. "
            "The attached handler will be ignored."
        ),
    ]

    def __init__(
        self,
        context: ContextData,
        client: httpx.AsyncClient,
        *,
        handler: AssertionHandler | None = None,
        handler_methodize: bool = False,
    ):
        self.context = context
        self.client = client
        self.handler = handler or AssertionHandler(context.config)
        self.handler_methodize = handler_methodize

    @functools.cached_property
    def context_wrapped(self) -> typer.Context:
        return typer.Context(Command("empty"), obj=self.context)

    @classmethod
    def spawn_from(cls, v: "BaseRequests") -> Self:
        "Make an instance from existing instance."
        return cls(
            context=v.context,
            client=v.client,
            handler=v.handler,
            handler_methodize=v.handler_methodize,
        )

    async def gather(
        self, *items: httpx.Request, **handler_kwargs
    ) -> Tuple[RequestHandlerData, ...]:
        responses = await asyncio.gather(*map(self.client.send, items))
        return self.handler(responses, **handler_kwargs)

    async def send(
        self, request: httpx.Request, **handler_kwargs
    ) -> RequestHandlerData:
        response = await self.client.send(request)
        return self.handler(response, **handler_kwargs)[0]

    def create_handler_args(self) -> Dict[str, Any]:
        return dict()


def params(**kwargs) -> Dict[str, Any]:
    return {k: v for k, v in kwargs.items() if v is not None}



# =========================================================================== #
import json
from datetime import datetime
from functools import cached_property
from typing import Annotated, List, Optional

import typer
import yaml
from fastapi import HTTPException
from pydantic import BaseModel
from rich.align import Align
from rich.console import Console
from rich.json import JSON
from rich.panel import Panel
from rich.table import Table
from sqlalchemy import delete, func, select
from sqlalchemy.orm import sessionmaker as _sessionmaker

# --------------------------------------------------------------------------- #
from captura import User
from captura.fields import KindObject
from captura.models import Base, Document
from captura.schemas import TimespanLimitParams, UserExtraSchema, mwargs
from legere.config import Output, OutputConfig
from legere.handlers import HandlerData
from legere.requests.base import BaseTyperizable, typerize
from simulatus import DummyHandler, DummyProvider, DummyProviderYAML
from simulatus.config import ConfigSimulatus, DummyConfig
from simulatus.reports import Report, ReportController, ReportView, row2dict

CONSOLE = Console()

# --------------------------------------------------------------------------- #
# Flags

FlagNote = Annotated[
    str,
    OPTION_NOTE := typer.Option(
        "--note",
        help="Note associated with a report.",
    ),
]
FlagNoteOptional = Annotated[Optional[str], OPTION_NOTE]
FlagBefore = Annotated[Optional[int], typer.Option("--before")]
FlagAfter = Annotated[Optional[int], typer.Option("--after")]
FlagLimit = Annotated[
    int,
    typer.Option(
        "--limit",
        "-n",
        help="Maximum number of reports to display.",
    ),
]
FlagTags = Annotated[Optional[List[str]], typer.Option("--tag")]
FlagUUIDUserOptional = Annotated[
    Optional[str],
    typer.Option(
        "--uuid-user",
        help="UUID of the report subject. When this is ``None``, aggregate reports are created.",
    ),
]
FlagUUIDUserListOptional = Annotated[
    Optional[List[str]], typer.Option("--uuid-user", help="User uuids.")
]

FlagCount = Annotated[
    int, typer.Option("--count", help="Number of dummies to generate/taint.")
]
FlagManifest = Annotated[
    Optional[str],
    typer.Option(
        "--manifest",
        "-f",
        help="Dummy configuration to overwrite that specified by the config.",
    ),
]
FlagManifestPreview = Annotated[
    bool,
    typer.Option("--preview", help="Manifest preview (as YAML)."),
]

ArgUUIDReport = Annotated[str, typer.Argument(help="Report UUID.")]


class ContextDataDummy(BaseModel):
    # NOTE: Global options for flags do not exist. Instead, create a manifest.
    quiet: bool = True
    config: ConfigSimulatus
    config_output: OutputConfig

    def register_manifest(self, manifest_path: str | None) -> DummyConfig | None:
        if manifest_path is None:
            return None

        manifest = DummyConfig.load(manifest_path)
        self.config.dummy = manifest

        return manifest

    def preview_manifest(self):
        data = HandlerData(
            data=self.config.dummy.model_dump(),
            output_config=self.config_output,
        )
        data.print()

    @cached_property
    def dummy_handler(self) -> DummyHandler:
        engine = self.config.engine()
        sm = _sessionmaker(engine)

        return DummyHandler(sm, self.config)

    @classmethod
    def for_typer(
        cls,
        context: typer.Context,
        quiet: Annotated[bool, typer.Option("--quiet/--loud")] = True,
        path_config: Annotated[Optional[str], typer.Option("--config")] = None,
    ) -> None:
        if path_config is None:
            config = mwargs(ConfigSimulatus)
        else:
            with open(path_config, "r") as file:
                config = ConfigSimulatus.model_validate(
                    yaml.safe_load(file),
                )

        config_output = mwargs(OutputConfig, output=Output.yaml)
        context.obj = cls(config=config, quiet=quiet, config_output=config_output)


# --------------------------------------------------------------------------- #
# Typerizables


class CmdReport(BaseTyperizable):
    """Reports management."""

    typer_decorate = False
    typer_check_verbage = False
    typer_commands = dict(
        user="user",
        aggregate="aggregate",
        view="view",
        history="history",
        delete="delete",
        all="all",
        amend="amend",
        destroy="destroy",
    )

    @classmethod
    def destroy(cls, _context: typer.Context):
        "Empty the reports table."
        context: ContextDataDummy = _context.obj
        with context.dummy_handler.sessionmaker() as session:
            session.execute(delete(Report))
            session.commit()

    @classmethod
    def delete(cls, _context: typer.Context, uuid_report: ArgUUIDReport):
        """Delete one report."""

        context: ContextDataDummy = _context.obj
        with context.dummy_handler.sessionmaker() as session:
            try:
                ReportView.delete_report(
                    ReportController(session),
                    uuid_report,
                )
            except HTTPException as err:
                print(err)
                raise typer.Exit(1)

    @classmethod
    def view(cls, _context: typer.Context, uuid_report: ArgUUIDReport):
        """View *(in detail)* a report."""

        context: ContextDataDummy = _context.obj
        with context.dummy_handler.sessionmaker() as session:
            try:
                report = ReportView.get_report(ReportController(session), uuid_report)
            except HTTPException as err:
                print(err)
                raise typer.Exit(1)

        CONSOLE.print(report)

    @classmethod
    def amend(
        cls,
        _context: typer.Context,
        uuid_report: ArgUUIDReport,
        note: FlagNoteOptional = None,
        tags: FlagTags = None,
    ):
        """Update a note on a report."""

        context: ContextDataDummy = _context.obj
        with context.dummy_handler.sessionmaker() as session:
            try:
                ReportView.put_report(
                    ReportController(session),
                    uuid_report,
                    note,
                    tags,  # type: ignore
                )
            except HTTPException as err:
                print(err)
                raise typer.Exit(1)

    @classmethod
    def history(
        cls,
        _context: typer.Context,
        uuid_user: FlagUUIDUserOptional = None,
        before: FlagBefore = None,
        after: FlagAfter = None,
        limit: FlagLimit = 10,
        kind_count: Annotated[
            Optional[KindObject],
            typer.Option("--kind-count", "-k", help="Order by this count"),
        ] = None,
        kind_count_desc: Annotated[
            bool,
            typer.Option(
                "--kind-count-desc/--kind-count-asc",
                help="Order direction of `--kind-count`.",
            ),
        ] = False,
        exclude_children: bool = True,
        tags: FlagTags = None,
    ):
        """View the latest reports."""

        context: ContextDataDummy = _context.obj
        tsp = mwargs(TimespanLimitParams, before=before, after=after, limit=limit)
        with context.dummy_handler.sessionmaker() as session:
            reports = ReportView.get_reports(
                ReportController(session),
                tsp,
                uuid_user,
                exclude_children=exclude_children,
                kind_count=kind_count,
                kind_count_desc=kind_count_desc,
                tags=tags,
            )

            if not reports:
                CONSOLE.print("[green]No reports to show.")
                raise typer.Exit(0)

            table = Table()
            for count, item in enumerate(reports):
                flattened = item.flatten()
                if count == 0:
                    tuple(map(table.add_column, flattened.keys()))

                style = "blue" if count % 2 else "cyan"
                table.add_row(*map(str, flattened.values()), style=style)

            CONSOLE.print(table)

    @classmethod
    def all(
        cls,
        _context: typer.Context,
        note: FlagNoteOptional = None,
        tags: FlagTags = None,
    ):
        note = note or "From `CmdSnapshot.all`."
        context: ContextDataDummy = _context.obj
        with context.dummy_handler.sessionmaker() as session:
            if not context.quiet:
                CONSOLE.print("[green]Building aggregate and user reports.")

            uuid_report = ReportView.post_report_build_all(
                ReportController(session),
                note,
                tags=tags,
            )

        if not context.quiet:
            CONSOLE.print("report_uuid", uuid_report)

    @classmethod
    def user(
        cls,
        _context: typer.Context,
        uuid: FlagUUIDUserOptional = None,
        note: FlagNoteOptional = None,
        tags: FlagTags = None,
    ):
        context: ContextDataDummy = _context.obj
        with context.dummy_handler.sessionmaker() as session:
            if uuid is not None:
                q = select(User).where(User.uuid == uuid)
                user = session.scalar(q)
                if user is None:
                    CONSOLE.print(f"No user with uuid `{uuid}`.")
                    raise typer.Exit(1)
            else:
                dummy_provider = DummyProvider(
                    context.dummy_handler.config,
                    session,
                    use_existing=False,
                )
                user = dummy_provider.user

            report = ReportView.post_report_build(
                ReportController(session),
                note=note or "From `CmdSnapshot.user`.",
                uuid_user=user.uuid,
                return_report=context.quiet,
                tags=tags,
            )

        if not context.quiet:
            CONSOLE.print(report)

    @classmethod
    def aggregate(
        cls,
        _context: typer.Context,
        note: FlagNoteOptional = None,
        tags: FlagTags = None,
    ):
        context: ContextDataDummy = _context.obj
        with context.dummy_handler.sessionmaker() as session:
            report = ReportView.post_report_build(
                ReportController(session),
                note=note or "From `CmdSnapshot.aggregate`.",
                uuid_user=None,
                return_report=not context.quiet,
                tags=tags,
            )

        if not context.quiet:
            CONSOLE.print(report)


class CmdUser(BaseTyperizable):
    """Manage dummy users."""

    typer_decorate = False
    typer_check_verbage = False
    typer_commands = dict(new="new", get="get", taint="taint", search="search")

    @classmethod
    def new(cls, _context: typer.Context, count: FlagCount = 1):
        """Spawn one (or more when ``--count`` is specified)."""
        context: ContextDataDummy = _context.obj
        handler = context.dummy_handler
        with handler.sessionmaker() as session:
            for _ in range(count):
                DummyProvider(handler.config, session, use_existing=False)

    @classmethod
    def get(cls, _context: typer.Context, count: FlagCount = 1):
        """Get one (or more when ``--count`` is specified). The intended use
        of this is as a tool to debug any strangeness with pytest fixtures.
        """
        context: ContextDataDummy = _context.obj
        handler = context.dummy_handler
        with handler.sessionmaker() as session:
            q_existing = select(User).order_by(func.random()).limit(count)
            existing = session.scalars(q_existing)
            for ee in existing:
                DummyProvider(handler.config, session, use_existing=ee)

    @classmethod
    def search(
        cls,
        _context: typer.Context,
        limit: FlagLimit = 10,
    ):
        context: ContextDataDummy = _context.obj
        q = DummyHandler.q_select(limit)
        with context.dummy_handler.sessionmaker() as session:
            table = Table()
            for count, row in enumerate(session.execute(q).all()):
                flattened = row2dict(row)
                if count == 0:
                    tuple(map(table.add_column, flattened.keys()))

                style = "blue" if count % 2 else "cyan"
                table.add_row(*map(str, flattened.values()), style=style)

            CONSOLE.print(table)

    @classmethod
    def taint(
        cls,
        _context: typer.Context,
        count: FlagCount = 1,
        uuid: FlagUUIDUserOptional = None,
    ):
        """Taint one (or more when ``--count`` is specified)."""
        context: ContextDataDummy = _context.obj
        handler = context.dummy_handler
        with handler.sessionmaker() as session:
            if uuid is not None:
                q = select(User).where(User.uuid == uuid)
            else:
                q = select(User).order_by(func.random()).limit(count)

            users = tuple(session.scalars(q))
            for user in users:
                DummyProvider(
                    context.config,
                    session,
                    use_existing=user,
                ).info_mark_tainted()

            dumped = json.dumps(
                [
                    UserExtraSchema.model_validate(user).model_dump(
                        mode="json", exclude={"name", "description"}
                    )
                    for user in users
                ]
            )
            session.commit()

            if not context.quiet:
                CONSOLE.print(Panel(Align.left(JSON(dumped))))


class CmdDummy(BaseTyperizable):
    """Manage Captura dummy data and create reports."""

    typer_decorate = False
    typer_check_verbage = False
    typer_commands = dict(
        preview="preview",
        dispose="dispose",
        apply="apply",
        destroy="destroy",
        initialize="initialize",
    )
    typer_children = dict(reports=CmdReport, users=CmdUser)

    @classmethod
    def dispose(
        cls,
        _context: typer.Context,
        manifest: FlagManifest = None,
        uuids: FlagUUIDUserListOptional = None,
        preview: FlagManifestPreview = False,
        prune: int = 0,
    ):
        """Dispose of all tainted dummies. To clear the database, see
        ``destroy``.
        """
        context: ContextDataDummy = _context.obj
        context.register_manifest(manifest)
        if prune:
            if prune < 2:
                CONSOLE.print("[red]Maximum uses too small to prune.")
                raise typer.Exit()

            context.config.dummy.users.maximum_uses = prune
        if preview:
            context.preview_manifest()
            return

        CONSOLE.print("[grenn]Cleaning up tainted data.")
        context.dummy_handler.dispose(set(uuids) if uuids is not None else None)
        CONSOLE.print("[green]Done cleaning up.")
        return

        # with context.dummy_handler.sessionmaker() as session:
        #     users = session.scalars(context.dummy_handler.q_clean())
        #     dumped = json.dumps(
        #         [
        #             UserExtraSchema.model_validate(user).model_dump(
        #                 mode="json", include={"uuid", "content"}
        #             )
        #             for user in users
        #         ]
        #     )
        #
        # CONSOLE.print(Panel(Align.left(JSON(dumped))))

    @classmethod
    def apply(
        cls,
        _context: typer.Context,
        manifest: FlagManifest = None,
        preview: FlagManifestPreview = False,
    ):
        """Create dummies to meet the criteria specified in config. To view the
        config, see ``config``.
        """

        context: ContextDataDummy = _context.obj
        context.register_manifest(manifest)
        if preview:
            context.preview_manifest()
            return

        from rich.progress import Progress

        CONSOLE.print("[green]Restoring dummies...")
        with context.dummy_handler.sessionmaker() as session:
            DummyProviderYAML.merge(session)

        start = datetime.now()
        with Progress() as progress:
            t = progress.add_task(description="Dummies Generated")
            context.dummy_handler.restore(
                callback=lambda dd, count, n: progress.update(
                    t,
                    advance=1,
                    total=n,
                )
            )
        end = datetime.now()

        CONSOLE.print("[green]Done restoring dummies.")
        CONSOLE.print(f"[green]Total time: {end - start}")

    @classmethod
    def initialize(cls, _context: typer.Context):
        """Add the tables required for reports to the database."""

        context: ContextDataDummy = _context.obj

        with context.dummy_handler.sessionmaker() as session:
            Base.metadata.create_all(session.bind)  # type: ignore[arg-type]
            # DummyProviderYAML.merge(session)

    @classmethod
    def destroy(
        cls,
        _context: typer.Context,
        destroy_reports: Annotated[
            bool,
            typer.Option(
                "--destroy-reports",
                help="When true, the reports table is emptied.",
            ),
        ] = False,
    ):
        """Empty all tables."""

        context: ContextDataDummy = _context.obj

        with context.dummy_handler.sessionmaker() as session:
            # NOTE: Using delete directly to avoid memory overhead. Relations
            #       are configured on the database side and not on this side
            #       using ``FOREIGN KEY ... ON DELETE CASCADE``.
            session.execute(delete(User))
            session.execute(delete(Document))
            if destroy_reports:
                session.execute(delete(Report))

            session.commit()

    @classmethod
    def preview(cls, _context: typer.Context, manifest_path: FlagManifest = None):
        context: ContextDataDummy = _context.obj
        context.register_manifest(manifest_path)
        context.preview_manifest()


# =========================================================================== #
import logging  # noqa: E402
import logging.config  # noqa: E402

logging.config.dictConfig


def main():
    tt = typerize(CmdDummy, callback=ContextDataDummy.for_typer)
    tt()


if __name__ == "__main__":
    main()



# =========================================================================== #
from typing import ClassVar

import httpx
import typer

# --------------------------------------------------------------------------- #
from captura.models import LevelStr
from legere import flags
from legere.handlers import AssertionHandler
from legere.requests.base import BaseRequests, ContextData, methodize, params


# NOTE: For management of document grants. Notice the duality between the
#       command names. This will be put on document requests, so it will be
#       used like `client documents grants ...`.
class DocumentGrantRequests(BaseRequests):
    @classmethod
    def req_read(
        cls,
        _context: typer.Context,
        uuid_document: flags.ArgUUIDDocument,
        *,
        level: flags.FlagLevel = LevelStr.view,
        uuid_user: flags.FlagUUIDUsersOptional = None,
        pending: flags.FlagPending = False,
        pending_from: flags.FlagPendingFromOptional = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "GET",
            context.url(cls.fmt_url.format(uuid_document)),
            params=params(
                uuid_user=uuid_user,
                level=level.name,
                pending=pending,
                pending_from=pending_from.name if pending_from is not None else None,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_invite(
        cls,
        _context: typer.Context,
        uuid_document: flags.ArgUUIDDocument,
        *,
        level: flags.FlagLevel = LevelStr.view,
        uuid_user: flags.FlagUUIDUsersOptional = None,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "POST",
            context.url(cls.fmt_url.format(uuid_document)),
            params=params(uuid_user=uuid_user, level=level.name, force=force),
            headers=context.headers,
        )

    @classmethod
    def req_revoke(
        cls,
        _context: typer.Context,
        uuid_document: flags.ArgUUIDDocument,
        *,
        uuid_user: flags.FlagUUIDUsers,
        force: flags.FlagForce = False,
        pending: flags.FlagPending = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "DELETE",
            context.url(cls.fmt_url.format(uuid_document)),
            headers=context.headers,
            params=dict(uuid_user=uuid_user, force=force, pending=pending),
        )

    @classmethod
    def req_approve(
        cls,
        _context: typer.Context,
        uuid_document: flags.ArgUUIDDocument,
        *,
        uuid_user: flags.FlagUUIDUsers,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "PATCH",
            context.url(cls.fmt_url.format(uuid_document)),
            params=params(uuid_user=uuid_user),
            headers=context.headers,
        )

    fmt_url: ClassVar[str] = "/grants/documents/{}"
    typer_check_verbage = False
    typer_commands = dict(
        read="req_read",
        invite="req_invite",
        revoke="req_revoke",
        approve="req_approve",
    )
    approve = methodize(req_approve, __func__=req_approve.__func__)  # type: ignore
    invite = methodize(req_invite, __func__=req_invite.__func__)  # type: ignore
    revoke = methodize(req_revoke, __func__=req_revoke.__func__)  # type: ignore
    read = methodize(req_read, __func__=req_read.__func__)  # type: ignore


class UserGrantRequests(BaseRequests):
    @classmethod
    def req_read(
        cls,
        _context: typer.Context,
        uuid_user: flags.ArgUUIDUser,
        *,
        level: LevelStr = LevelStr.view,
        uuid_document: flags.FlagUUIDDocumentsOptional = None,
        pending: flags.FlagPending = False,
        pending_from: flags.FlagPendingFromOptional = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "GET",
            context.url(cls.fmt_url.format(uuid_user)),
            params=params(
                level=level.name,
                uuid_document=uuid_document,
                pending=pending,
                pending_from=pending_from.name if pending_from is not None else None,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_accept(
        cls,
        _context: typer.Context,
        uuid_user: flags.ArgUUIDUser,
        *,
        uuid_document: flags.FlagUUIDDocumentsOptional = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "PATCH",
            context.url(cls.fmt_url.format(uuid_user)),
            params=params(uuid_document=uuid_document),
            headers=context.headers,
        )

    @classmethod
    def req_reject(
        cls,
        _context: typer.Context,
        uuid_user: flags.ArgUUIDUser,
        *,
        uuid_document: flags.FlagUUIDDocumentsOptional = None,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "DELETE",
            context.url(cls.fmt_url.format(uuid_user)),
            params=params(
                uuid_document=uuid_document,
                force=force,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_request(
        cls,
        _context: typer.Context,
        uuid_user: flags.ArgUUIDUser,
        *,
        level: LevelStr = LevelStr.view,
        uuid_document: flags.FlagUUIDDocumentsOptional = None,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "POST",
            context.url(cls.fmt_url.format(uuid_user)),
            params=params(
                level=level.name,
                uuid_document=uuid_document,
                force=force,
            ),
            headers=context.headers,
        )

    fmt_url: ClassVar[str] = "/grants/users/{}"
    typer_check_verbage = False
    typer_commands = dict(
        read="req_read",
        request="req_request",
        reject="req_reject",
        accept="req_accept",
    )

    read = methodize(req_read, __func__=req_read.__func__)  # type: ignore
    request = methodize(req_request, __func__=req_request.__func__)  # type: ignore
    reject = methodize(req_reject, __func__=req_reject.__func__)  # type: ignore
    accept = methodize(req_accept, __func__=req_accept.__func__)  # type: ignore


class GrantRequests(BaseRequests):
    typer_children = dict(
        documents=DocumentGrantRequests,
        users=UserGrantRequests,
    )
    documents: DocumentGrantRequests
    users: UserGrantRequests

    def __init__(
        self,
        context: ContextData,
        client: httpx.AsyncClient,
        *,
        handler: AssertionHandler | None = None,
        handler_methodize: bool = True,
        documents: DocumentGrantRequests | None = None,
        users: UserGrantRequests | None = None,
    ):
        super().__init__(
            context,
            client,
            handler=handler,
            handler_methodize=handler_methodize,
        )
        self.documents = documents or DocumentGrantRequests.spawn_from(self)
        self.users = users or UserGrantRequests.spawn_from(self)

    @property
    def d(self) -> DocumentGrantRequests:
        return self.documents

    @property
    def u(self) -> UserGrantRequests:
        return self.users


__all__ = (
    "DocumentGrantRequests",
    "UserGrantRequests",
    "GrantRequests",
)


if __name__ == "__main__":
    # --------------------------------------------------------------------------- #
    from legere.requests.base import typerize

    grants = typerize(GrantRequests)
    grants()



# =========================================================================== #
import base64
import hashlib
import json
import secrets
from datetime import datetime
from typing import Annotated, Any, Dict

import httpx
from fastapi import Depends, Form, HTTPException, Request
from fastapi.responses import RedirectResponse
from pydantic import ValidationError
from sqlalchemy import String, func, select
from sqlalchemy.orm import Mapped, mapped_column

# --------------------------------------------------------------------------- #
from captura import fields
from captura.auth import Token, TokenPermissionTier, try_decode
from captura.config import Config
from captura.controllers.base import Data, ResolvedUser
from captura.controllers.create import Create
from captura.controllers.read import Read
from captura.depends import DependsAuth, DependsConfig, DependsSessionMaker
from captura.models import Base, MappedColumnUUID, User
from captura.schemas import UserCreateSchema, mwargs
from captura.views.base import BaseView, OpenApiResponseCommon, OpenApiTags
from captura.views.users import UserView


def check_auth0_data(auth0_data: Dict[str, Any]):
    bad = set(v for v in AUTH0_KEYS if v not in auth0_data)
    if len(bad):
        detail = f"Missing expected fields `{bad}`."
        raise HTTPException(422, detail=detail)


def exclude_for(use_auth0: bool = True):
    mode = "auth0" if use_auth0 else "pytest"

    def wrapper(config: DependsConfig):
        if config.auth0.use is use_auth0:
            raise HTTPException(
                409,
                detail=f"Not available in `{mode}` mode.",
            )

    return wrapper


AUTH0_KEY_TOKEN_ACCESS = "access_token"
AUTH0_KEY_TOKEN_ID = "id_token"
AUTH0_KEYS = {AUTH0_KEY_TOKEN_ID, AUTH0_KEY_TOKEN_ACCESS}
DEMO_KEY_UUID = "demo_uuid"
DEMO_KEY_EXISTS_USER = "demo_exists_user"
DEMO_KEY_EXISTS = "demo_exists"


def form_body(cls):
    cls.__signature__ = cls.__signature__.replace(
        parameters=[
            arg.replace(default=Form(...))
            for arg in cls.__signature__.parameters.values()
        ]
    )
    return cls


def token_session(request: Request, auth: DependsAuth) -> Token:
    if (token := request.session.get(AUTH0_KEY_TOKEN_ACCESS)) is None:
        raise HTTPException(401, detail="Token required.")

    return Token.decode(auth, token, header=False)


DependsTokenSession = Annotated[Token, Depends(token_session)]


class AuthViewPytest(BaseView):
    """This is where routes to handle login and getting tokens will be."""

    view_routes = dict(
        post_register=dict(
            url="/register",
            name="",
            dependencies=[Depends(exclude_for(use_auth0=True))],
        ),
        post_token=dict(
            url="/token",
            name="Mint Test Token",
            dependencies=[Depends(exclude_for(use_auth0=True))],
        ),
        get_token=dict(
            url="/token",
            name="Verify Token",
        ),
    )
    view_router_args = dict(
        tags=[OpenApiTags.auth0],
        responses=OpenApiResponseCommon,
    )

    @classmethod
    def post_register(
        cls,
        sessionmaker: DependsSessionMaker,
        auth: DependsAuth,
        email: str,
        name: str,
        description: str,
        url_image: str | None = None,
        url: str | None = None,
    ) -> str:
        """Create user, return token."""

        try:
            registration_data = mwargs(
                UserCreateSchema,
                name=name,
                description=description,
                url_image=url_image,
                url=url,
                email=email,
            )
        except ValidationError as err:
            raise HTTPException(422, detail=err.json())

        with sessionmaker() as session:
            q = select(User.uuid).where(User.email == email)
            email_exists = session.scalar(q) is not None
            if email_exists:
                raise HTTPException(400, detail="Account with email already exists.")

            create = Create[UserCreateSchema](
                session,
                token=None,
                method="POST",
                api_origin="POST /register",
            )
            create.create_data = registration_data
            data: Data[ResolvedUser]
            data = mwargs(
                Data[ResolvedUser],
                token_user=None,
                data=ResolvedUser.empty(),
            )
            data_final = create.user(data)
            (user,) = data_final.data.users
            user.subject = hashlib.sha256(user.uuid.encode()).hexdigest()
            data_final.commit(create.session)

        token_final = mwargs(
            Token,
            sub=data_final.data.users[0].uuid,
            tier=TokenPermissionTier.paid,
        )
        return token_final.encode(auth)

    @classmethod
    def post_token(cls, auth: DependsAuth, data: Token) -> str:
        """Use this to create a new token.

        This endpoint only works when authentication is in pytest mode, and
        will not use auth0 mode.
        """

        return data.try_encode(auth)

    @classmethod
    def get_token(cls, auth: DependsAuth, data: str) -> Token:
        """Use this to validate JWT payload. Decodes and returns."""

        decoded, err = try_decode(auth, data, header=False)
        if err is not None:
            raise err

        assert decoded is not None
        # decoded["uuid"] = "sub"
        return Token.model_validate(decoded)


# NOTE: This is starting to follow the extension pattern. I might refactor it
#       as such later. At the moment this table can be created by running
#       ``simulatus initialize``
class Demo(Base):
    __tablename__ = "demos"

    uuid: Mapped[MappedColumnUUID] = mapped_column(primary_key=True)
    email: Mapped[str] = mapped_column(String(fields.LENGTH_NAME), unique=True)
    timestamp: Mapped[int] = mapped_column(
        default=(_now := lambda: datetime.timestamp(datetime.now())),
    )


class AuthViewAuth0(BaseView):

    view_routes = {
        "get_login": {
            "url": "/login",
            "tags": [OpenApiTags.auth0],
            "name": "Login Via Auth0",
            "description": "Login using auth0 and recieve a JWT.",
        },
        "get_logout": {
            "url": "/logout",
            "tags": [OpenApiTags.auth0],
            "name": "Logout",
        },
        "get_authorize": {
            "url": "/authorize",
            "tags": [OpenApiTags.auth0],
            "name": "Login Callback",
            "description": "Get a JSON web token.",
        },
        "get_register": {
            "url": "/register",
            "tags": [OpenApiTags.auth0],
            "name": "Register",
            "description": "Register a new user.",
        },
        "post_register": {
            "url": "/register",
            "tags": [OpenApiTags.auth0],
            "name": "Register",
            "description": "Register a new user.",
        },
        "get_profile": {
            "url": "/profile",
            "tags": [OpenApiTags.auth0],
            "name": "Profile",
            "description": "Profile.",
        },
        "get_demo": {
            "url": "/demo",
            "tags": [OpenApiTags.auth0],
            "name": "Request a Demo",
            "description": "Request a demo.",
        },
        "post_demo": {
            "url": "/demo",
            "tags": [OpenApiTags.auth0],
            "name": "Handle Demo Request.",
            "description": "Handle a demo request.",
        },
    }

    view_router_args = dict(dependencies=[Depends(exclude_for(use_auth0=False))])

    # NOTE: Decided to do this with httpx and redirects since I only want to
    #       implement code exchange flow (RFC 6749,
    #       see https://datatracker.ietf.org/doc/html/rfc6749#section-4.1).
    #       Further, it is worth mentioning that I tried using ``authlib``
    #       as suggested by the quickstart but debugging token audience issues
    #       was a huge pain in the ass and it ultimately made more sense to
    #       just use httpx.
    @classmethod
    async def get_login(cls, config: DependsConfig):
        url_redirect = f"{config.app.host_url}/authorize"

        auth0 = config.auth0
        params = f"?response_type=code&client_id={auth0.app.client_id}"
        params += "&scope=openid profile email"
        params += f"&state={secrets.token_urlsafe(8)}"
        params += f"&redirect_uri={url_redirect}"
        params += f"&audience={config.auth0.api.audience[0]}"
        url_auth0 = f"https://{config.auth0.issuer}/authorize{params}"

        return RedirectResponse(url_auth0)

    @classmethod
    async def get_logout(cls, request: Request):
        request.session.clear()
        return RedirectResponse("/", 302)

    @classmethod
    async def get_authorize(
        cls,
        request: Request,
        config: DependsConfig,
        sessionmaker: DependsSessionMaker,
        auth: DependsAuth,
        code: str,
    ):
        """Exchange the code for a token.

        Redirects new users to registration and existing users to their
        profile.
        """

        response = httpx.post(
            f"https://{config.auth0.issuer}/oauth/token",
            data=dict(
                redirect_uri=f"{config.app.host_url}/authorize",
                code=code,
                grant_type="authorization_code",
                # audience=config.auth0.api.audience,
                client_id=config.auth0.app.client_id,
                client_secret=config.auth0.app.client_secret.get_secret_value(),
            ),
        )
        if response.status_code != 200:
            detail = "Failed to exchange code for token."
            raise HTTPException(500, detail=detail)

        # NOTE: Ensure that the auth0 response is as expected.
        try:
            auth0_data = response.json()
        except json.JSONDecodeError as err_json:
            detail = "Failed to decode Auth0 response."
            raise HTTPException(500, detail=detail) from err_json

        # NOTE: Check the browser cookie and decode the token to get the
        #       subject.
        check_auth0_data(auth0_data)
        decoded, err = try_decode(
            auth, auth0_data[AUTH0_KEY_TOKEN_ACCESS], header=False
        )
        if err is not None:
            raise err

        assert decoded is not None
        subject_256 = hashlib.sha256(decoded["sub"].encode()).hexdigest()

        # NOTE: See RFC 6648. Make believe header to pass info on since query
        #       parameters are insecure (because of logging, etc).
        #       See further https://stackoverflow.com/questions/3561381/custom-http-headers-naming-conventions
        request.session.update({key: auth0_data[key] for key in AUTH0_KEYS})

        with sessionmaker() as session:
            q = select(func.count(User.uuid)).where(User.subject == subject_256)
            n = session.scalar(q)

        url = "/profile" if n else "/register"
        return RedirectResponse(url, status_code=302)

    # ----------------------------------------------------------------------- #
    # Registration

    @classmethod
    def check_code(cls, config: Config, email: str, code: str) -> None:

        # NOTE: Code to expect should be the sum of the salt and the users
        #       email address.
        code_expect = cls.create_code(config, email)
        if code != code_expect:
            raise HTTPException(403, detail="Invalid code.")

    @classmethod
    def create_code(cls, config: Config, email: str):
        return hashlib.sha256(
            config.auth0.registration_code_salt + email.encode()
        ).hexdigest()

    @classmethod
    async def get_register(
        cls,
        config: DependsConfig,
        auth: DependsAuth,
        request: Request,
    ):
        """This should return a form containing user creation parameters.

        Submission of this form should send `POST /register` with this data
        and then redirect the user to `/profile`
        """

        check_auth0_data(request.session)
        # access_token_raw = request.session[AUTH0_KEY_TOKEN_ACCESS]
        if AUTH0_KEY_TOKEN_ACCESS not in request.session:
            raise HTTPException(403, "Session missing `access_token`.")

        id_token_raw = request.session[AUTH0_KEY_TOKEN_ID]

        if len(id_token_split := id_token_raw.split(".")) == 3:
            _deco = base64.b64decode(id_token_split[1] + "==")
            id_token_decoded = json.loads(_deco)
        else:
            id_token_decoded = "Failed to decode `id_token`."

        # NOTE: Populate defaults from the id token.
        return cls.view_templates.TemplateResponse(
            request,
            name="register.j2",
            context=dict(id_token=id_token_decoded),
        )

    @classmethod
    async def post_register(
        cls,
        request: Request,
        auth: DependsAuth,
        sessionmaker: DependsSessionMaker,
        config: DependsConfig,
        code: str = Form(),
        email: str = Form(),
        name: str = Form(),
        description: str = Form(),
        url_image: str | None = Form(None),
        url: str | None = Form(None),
    ):
        """Accepts form data. Creates user. Redirects to profile."""

        check_auth0_data(request.session)
        access_token = auth.decode(
            request.session[AUTH0_KEY_TOKEN_ACCESS], header=False
        )
        subject = access_token["sub"]
        subject_256 = hashlib.sha256(subject.encode()).hexdigest()

        cls.check_code(config, email, code)
        try:
            registration_data = UserCreateSchema(
                content={},
                name=name,
                description=description,
                url_image=url_image,
                url=url,
                email=email,
            )
        except ValidationError as err:
            raise HTTPException(422, detail=err.json())

        with sessionmaker() as session:
            q = select(User.uuid).where(User.email == email)
            email_exists = session.scalar(q) is not None
            if email_exists:
                raise HTTPException(400, detail="Account with email already exists.")

            create = Create[UserCreateSchema](
                session,
                token=None,
                method="POST",
                api_origin="POST /register",
            )
            create.create_data = registration_data
            data: Data[ResolvedUser]
            data = mwargs(
                Data[ResolvedUser],
                token_user=None,
                data=ResolvedUser.empty(),
            )
            data_final = create.user(data)
            (user,) = data_final.data.users
            user.subject = subject_256
            data_final.commit(create.session)

        return RedirectResponse("/profile", 302)

    # ----------------------------------------------------------------------- #
    @classmethod
    async def get_demo(cls, request: Request):
        """Request a demo."""

        # NOTE: Maybe wrap in context instead?
        logged_in = request.session.get(AUTH0_KEY_TOKEN_ACCESS) is not None
        uuid = request.session.get(DEMO_KEY_UUID)
        exists = request.session.get(DEMO_KEY_EXISTS)
        exists_user = request.session.get(DEMO_KEY_EXISTS_USER)

        # NOTE: Says waiting for demo if already requested otherwises gets
        #       email and posts.
        return cls.view_templates.TemplateResponse(
            request,
            name="demo.j2",
            context=dict(
                uuid=uuid,
                logged_in=logged_in,
                exists=exists,
                exists_user=exists_user,
            ),
        )

    @classmethod
    async def post_demo(
        cls,
        request: Request,
        sessionmaker: DependsSessionMaker,
        email: Annotated[str, Form()],
    ) -> RedirectResponse:
        """Process demo data."""

        with sessionmaker() as session:
            # NOTE: Check if there is already a user with this email.
            user = session.scalar(select(User).where(User.email == email))
            if user is not None:
                request.session[DEMO_KEY_EXISTS_USER] = True
                return RedirectResponse(
                    "/demo",
                    status_code=302,
                )

            # NOTE: Check if demo exists.
            demo = session.scalar(select(Demo).where(Demo.email == email))
            exists = demo is not None
            if demo is None:
                demo = Demo(email=email)
                session.add(demo)
                session.commit()
                session.refresh(demo)

        request.session[DEMO_KEY_UUID] = demo.uuid
        request.session[DEMO_KEY_EXISTS] = exists
        request.session[DEMO_KEY_EXISTS_USER] = False

        return RedirectResponse(
            "/demo",
            status_code=302,
        )

    # NOTE: Save demo request somewhere.

    # ----------------------------------------------------------------------- #

    @classmethod
    async def get_profile(
        cls,
        request: Request,
        sessionmaker: DependsSessionMaker,
        token: DependsTokenSession,
    ):

        with sessionmaker() as session:
            read = Read(session, token, method="GET")
            user = token.validate_db(session)
            user_data = UserView.get_user(user.uuid, read)

        return cls.view_templates.TemplateResponse(
            request,
            name="profile.j2",
            context=dict(
                data=user_data,
                access_token=request.session[AUTH0_KEY_TOKEN_ACCESS],
            ),
        )



# =========================================================================== #
import json
from http import HTTPMethod
from typing import Any, ClassVar, Generic, Iterable, Protocol, Tuple, Type, TypeVar

import httpx
import yaml
from pydantic import BaseModel, TypeAdapter, ValidationError
from rich.align import Align, AlignMethod, VerticalAlignMethod
from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

# --------------------------------------------------------------------------- #
from captura.err import ErrDetail
from captura.schemas import mwargs
from legere.config import Config, OutputConfig
from legere.flags import Output

# Types and Constants
# About `CONSOLE_THEME`: https://pygments.org/docs/styles/#getting-a-list-of-available-styles

CONSOLE_THEME = "fruity"
CONSOLE = Console()


# NOTE: Old table stuff.
# def print_table(self, res: httpx.Response, data=None) -> int:
#     data = res.json() or data
#     table = self.create_table(res, data)
#     if table is None:
#         return 1
#     CONSOLE.print(table)
#     return 0
#
# def create_table(self, res: httpx.Response, data=None) -> Table | None:
#     table = Table()
#     if not (data := self.unfuck(data or res.json())):
#         CONSOLE.print("[green]No data to display.")
#         return None
#
#     # Create columns
#     keys = tuple(data[0].keys())
#     if self.columns:
#         keys = tuple(key for key in self.columns if keys)
#
#     if include_count := not self.columns or "count" in self.columns:
#         table.add_column("count")
#
#     column_config_base = dict(
#         justify="center",
#     )
#
#     for ii, key in enumerate(keys):
#         column_config_extra = self.column_configs.get(key, dict())
#         column_config: Dict[str, Any]
#         column_config = column_config_base | column_config_extra
#
#         if column_config.get("style") is None:
#             style = typer.colors.CYAN if not (ii % 2) else typer.colors.BLUE
#             column_config.update(style=style)
#
#         table.add_column(key, **column_config)
#
#     # Add rows
#     def omit(row: Dict[str, Any], row_last: Dict[str, Any], key: str):
#         value_row_last, value_row = row_last.get(key), row.get(key)
#         if value_row == value_row_last:
#             return "`"
#         if isinstance(value_row, list):
#             value_row = ", ".join(value_row)
#         return str(value_row)
#
#     row_last: Dict[str, Any] = dict()
#     for count, row in enumerate(data):
#         flat = tuple(omit(row, row_last, key) for key in keys)
#         if include_count:
#             table.add_row(str(count), *flat)
#         else:
#             table.add_row(*flat)
#         row_last = row
#
#     return table
#
# def unfuck(self, data):
#     if isinstance(data, dict):
#         if not all(not isinstance(item, dict) for item in data):
#             data = [data]
#         else:
#             data = list(
#                 dict(uuid=uuid, **item) if "uuid" not in item else item
#                 for uuid, item in data.items()
#             )
#     if not isinstance(data, list):
#         msg = "Results must be a coerced into a `list` "
#         msg += f"(is `{type(data)}`)."
#         raise HandlerError(msg)
#
#     return data

T_HandlerData = TypeVar("T_HandlerData")


class BaseHandlerData(Generic[T_HandlerData]):

    data: Any
    output_config: OutputConfig
    adapter: TypeAdapter[T_HandlerData] | None

    def __init__(
        self,
        data: T_HandlerData | None = None,
        output_config: OutputConfig | None = None,
        adapter: TypeAdapter[T_HandlerData] | None = None,
    ) -> None:
        self.data = data
        self.adapter = adapter
        self.output_config = output_config or mwargs(OutputConfig)

    def __rich__(self) -> Layout | Panel | str:
        render, _ = self.render()
        return render

    def __str__(self):
        return self.render_raw(output=Output.json)

    def print(self) -> int:
        rendered, exit_code = self.render()
        if self.output_config.output == Output.raw and isinstance(rendered, str):
            print(rendered)
        else:
            CONSOLE.print(rendered)
        return exit_code

    def data_jsonable(self, data: Any | None = None) -> Any:
        data = data or self.data
        if self.adapter is not None:
            return self.adapter.dump_python(data)

        return data

    def render(
        self,
        *,
        data: Any | None = None,
        output: Output | None = None,
        decorate: bool | None = None,
    ) -> Tuple[Any, int]:

        # NOTE: Print both the request and response. This will be used in
        #       pytest too.
        rendered: Any
        match (output or self.output_config.output):
            case Output.json:
                rendered = self.render_json(data=data, decorate=decorate)
            case Output.yaml:
                rendered = self.render_yaml(data=data, decorate=decorate)
            case Output.table:
                rendered = self.render_table(data=data)
            case Output.raw:
                rendered = self.render_raw(
                    data=data,
                    output=Output.json,
                )
            case _ as bad:
                raise ValueError(f"Unknown output format `{bad}`.")

        return rendered, 0

    def render_raw(
        self,
        *,
        data: Any | None = None,
        output: Output | None = None,
        **kwargs,
    ) -> str:
        match (output if output is not None else self.output_config.output):
            case Output.yaml:
                return "---\n" + yaml.dump(self.data_jsonable(data=data), **kwargs)
            case Output.json:
                return json.dumps(self.data_jsonable(data=data), **kwargs)
            case bad:
                raise ValueError(f"Cannot dump for kind `{bad}` in raw mode.")

    def render_json(
        self,
        align: AlignMethod = "left",
        vertical: VerticalAlignMethod = "middle",
        *,
        data: Any | None = None,
        decorate: bool | None = None,
    ) -> Panel:
        jj: Any = Syntax(
            self.render_raw(data=data, output=Output.json, indent=2),
            "json",
            theme=self.output_config.rich_theme,
            word_wrap=True,
            background_color="default",
        )
        jj = Align(jj, align, vertical=vertical)
        if decorate or self.output_config.decorate:
            jj = Panel(jj)
        return jj

    def render_yaml(
        self,
        align: AlignMethod = "left",
        vertical: VerticalAlignMethod = "middle",
        *,
        data: Any | None = None,
        decorate: bool | None = None,
    ) -> Panel:
        jj: Any = Syntax(
            self.render_raw(data=data, output=Output.yaml),
            "yaml",
            theme=self.output_config.rich_theme,
            word_wrap=True,
            background_color="default",
        )
        jj = Align(jj, align, vertical=vertical)
        if decorate or self.output_config.decorate:
            jj = Panel(jj)
        return jj

    def render_table(self, *, data: Any | None = None) -> Layout:
        raise ValueError("Table rendering is not available at this time.")


class HandlerData(BaseHandlerData[T_HandlerData]):
    """Use this to render data that has no request.

    This keeps the logic for requests entirely separate.
    """

    ...


class RequestHandlerData(BaseHandlerData[T_HandlerData]):

    response: httpx.Response
    expect_status: int
    expect_err: ErrDetail | None

    def __init__(
        self,
        response: httpx.Response,
        adapter: Type[BaseModel] | TypeAdapter[T_HandlerData] | None = None,
        *,
        data: T_HandlerData | None = None,
        output_config: OutputConfig | None = None,
        expect_err: ErrDetail | None = None,
        expect_status: int | None = None,
    ) -> None:
        self.output_config = output_config or mwargs(OutputConfig)
        self.expect_err = expect_err

        adapter_final: TypeAdapter | None
        if adapter is not None and not isinstance(adapter, TypeAdapter):
            adapter_final = TypeAdapter(adapter)
        else:
            adapter_final = adapter

        self.set_data(response, adapter_final, data=data, expect_status=expect_status)

    def set_data(
        self,
        response: httpx.Response,
        adapter: TypeAdapter[T_HandlerData] | None = None,
        *,
        data: T_HandlerData | None = None,
        expect_status: int | None = None,
    ) -> None:

        self.response = response
        if expect_status is None:
            match response.request.method:
                case HTTPMethod.POST:
                    expect_status = 201
                case _:
                    expect_status = 200

        self.expect_status = expect_status

        match [adapter, data]:
            case [None, None]:
                try:
                    self.data = response.json() if response.content else None
                except json.JSONDecodeError:
                    self.data = None

                self.adapter = None
            case [TypeAdapter() as adptr, dd]:
                if data is None:
                    try:
                        self.data = adptr.validate_json(response.content)
                    except ValidationError:
                        msg = "Failed to handle response with adapter `%s`." % adptr
                        CONSOLE.print(msg)
                        self.data = response.json() if response.content else None
                else:
                    self._data = adptr.validate_python(dd)
                self.adapter = adptr
            case [None, dd] if dd is not None:
                self.data = dd
                self.adapter = None
            case _:
                raise ValueError("Cannot set from data.")

        return None

    def render(
        self,
        *,
        data: Any | None = None,
        output: Output | None = None,
        decorate: bool | None = None,
    ) -> Tuple[Any, int]:
        res = self.response
        if res.status_code != self.expect_status:
            return (self.render_err(), 1)
        return super().render(data=data, output=output, decorate=decorate)

    def render_err(self) -> Table | None:
        """Return the response as rendered by :meth:`render_request` only when
        the status code is bad."""

        if self.expect_status != self.response.status_code:
            table = self.render_response()
            table.title = Text("Request Failed", style="bold red")
            return table

        if self.expect_err is None:
            return None

        return self.expect_err.compare(self.response)

    def render_response(self) -> Table:
        """Does not necessarily have to render an error."""

        table = Table(show_header=False, box=None, pad_edge=False)
        table.add_column()
        table.add_column()

        table.add_row(
            Text("Response Status Expected:", style="bold red"),
            Text(str(self.expect_status), style="bold yellow"),
        )
        table.add_row(
            Text("Response Status Recieved:", style="bold red"),
            Text(str(self.response.status_code), style="bold yellow"),
        )

        if self.response.content:
            try:
                response_data = self.response.json()
            except json.JSONDecodeError:
                response_data = None
        else:
            response_data = None

        response_data_handler: HandlerData = HandlerData(
            output_config=self.output_config,
            data=response_data,
        )
        table.add_row(
            Text("Response Content:", style="bold red"),
            response_data_handler.render(output=Output.raw)[0],
        )
        table.add_row()
        table.add_row(
            Text("Response Headers:", style="bold red"),
            render_headers(self.response),
        )

        self.render_request(table=table)
        return table

    def render_request(self, table: Table | None = None):
        return render_request(
            self.output_config,
            self.response.request,
            table=table,
        )


def render_headers(has_headers: httpx.Response | httpx.Request) -> Table:

    table = Table(box=None, pad_edge=False, show_header=False)
    table.add_column(width=16)
    table.add_column(no_wrap=False, overflow="ellipsis")
    for key, value in has_headers.headers.items():
        table.add_row(key, value)

    return table


def render_request(
    output_config: OutputConfig,
    request: httpx.Request,
    table: Table | None = None,
    # max_content_length: int | None = 4096,
) -> Table:

    if table is None:
        table = Table(show_header=False, box=None, pad_edge=False)
    else:
        table.add_row()

    table.add_row(
        Text("Request Method:", style="bold green"),
        Text(str(request.method), "bold yellow"),
    )
    table.add_row(
        Text("Request URL:", style="bold green"),
        Text(str(request.url), style="bold yellow"),
    )

    response_data = json.loads(request.content) if request.content else None
    response_data_handler: HandlerData = HandlerData(
        output_config=output_config,
        data=response_data,
    )
    table.add_row(
        Text("Request Content:", style="bold green"),
        response_data_handler.render(output=Output.raw)[0],
    )
    table.add_row()
    table.add_row(
        Text("Request Headers:", style="bold green"),
        render_headers(request),
    )

    return table


T_RequestHandlerReturn = TypeVar("T_RequestHandlerReturn", covariant=True)
ResolvableRequestHandlerDatas = Tuple[RequestHandlerData, ...] | RequestHandlerData
ResolvableHandlerDatas = Tuple[HandlerData, ...] | HandlerData


class BaseRequestHandler(Protocol, Generic[T_RequestHandlerReturn]):
    console: ClassVar[Console] = CONSOLE
    config: Config

    def __init__(self, config: Config):
        self.config = config

    def __call__(
        self,
        res: httpx.Response | Iterable[httpx.Response] | None,
        data: Any | Tuple[Any, ...] | None = None,
        adapter: TypeAdapter | None = None,
        *,
        expect_err: ErrDetail | None = None,
        expect_status: int | None = None,
        request_handler_data: ResolvableRequestHandlerDatas | None = None,
        handler_data: ResolvableHandlerDatas | None = None,
    ) -> T_RequestHandlerReturn: ...

    def handle(
        self,
        response: httpx.Response | Iterable[httpx.Response] | None = None,
        data: Any | Tuple[Any, ...] | None = None,
        adapter: TypeAdapter | None = None,
        *,
        expect_err: ErrDetail | None = None,
        expect_status: int | None = None,
        request_handler_data: ResolvableRequestHandlerDatas | None = None,
        handler_data: ResolvableHandlerDatas | None = None,
    ) -> T_RequestHandlerReturn:
        return self(
            response,
            data,
            adapter,
            expect_err=expect_err,
            expect_status=expect_status,
            request_handler_data=request_handler_data,
            handler_data=handler_data,
        )

    def resolve(
        self,
        response: httpx.Response | Iterable[httpx.Response] | None = None,
        data: Any | Tuple[Any, ...] | None = None,
        adapter: TypeAdapter | None = None,
        *,
        expect_err: ErrDetail | None = None,
        expect_status: int | None = None,
        request_handler_data: ResolvableRequestHandlerDatas | None = None,
    ) -> Tuple[RequestHandlerData, ...]:

        request_handler_datas: Tuple[RequestHandlerData, ...]
        match (response, request_handler_data):
            case (httpx.Response() as response, None):
                responses = (response,)
            case ([*responses], None):
                ...
            case (None, RequestHandlerData() as hd):
                return (hd,)
            case bad:
                raise ValueError(
                    f"Must specify exactly one of `requests` and "
                    f"`request_handler_data`. Input = `{bad}`."
                )

        request_handler_datas = tuple(
            RequestHandlerData(
                response=rr,
                data=data,
                adapter=adapter,
                output_config=self.config.output,
                expect_err=expect_err,
                expect_status=expect_status,
            )
            for rr in responses
        )

        return request_handler_datas

    def collect(
        self, request_handler_datas: Tuple[RequestHandlerData, ...]
    ) -> Tuple[Table, ...]:

        return tuple(
            table
            for handler_data in request_handler_datas
            if (table := handler_data.render_err()) is not None
        )


class ConsoleHandler(BaseRequestHandler[int]):
    def __call__(
        self,
        response: httpx.Response | Iterable[httpx.Response] | None = None,
        data: Any | Tuple[Any, ...] | None = None,
        adapter: TypeAdapter | None = None,
        *,
        expect_err: ErrDetail | None = None,
        expect_status: int | None = None,
        request_handler_data: ResolvableRequestHandlerDatas | None = None,
        handler_data: ResolvableHandlerDatas | None = None,
    ) -> int:

        if request_handler_data is not None or response is not None:
            if handler_data:
                raise ValueError(
                    "Cannot specify `handler_data` when "
                    "`request_handler_data` can be resolved."
                )
            request_handler_datas = self.resolve(
                response,
                data,
                adapter,
                expect_err=expect_err,
                expect_status=expect_status,
                request_handler_data=request_handler_data,
            )

            exit_code = 0
            for rhd in request_handler_datas:
                exit_code += rhd.print()

            return exit_code

        assert handler_data is not None
        if isinstance(handler_data, tuple):
            handler_datas = handler_data
        else:
            handler_datas = (handler_data,)

        exit_code = 0
        for hd in handler_datas:
            exit_code += hd.print()

        return exit_code


# TODO: Come back here!
class AssertionHandler(BaseRequestHandler[Tuple[RequestHandlerData, ...]]):

    def __call__(
        self,
        response: httpx.Response | Iterable[httpx.Response] | None = None,
        data: Any | Tuple[Any, ...] | None = None,
        adapter: TypeAdapter | None = None,
        *,
        expect_err: ErrDetail | None = None,
        expect_status: int | None = None,
        request_handler_data: ResolvableRequestHandlerDatas | None = None,
        handler_data: ResolvableHandlerDatas | None = None,
    ) -> Tuple[RequestHandlerData, ...]:

        if handler_data is not None:
            msg = "`AssertionHandler` does not accept `handler_data`."
            raise ValueError(msg)

        request_handler_datas, err = self.check_status(
            response,
            data,
            adapter,
            expect_err=expect_err,
            expect_status=expect_status,
            request_handler_data=request_handler_data,
        )
        if err is not None:
            raise err

        return request_handler_datas

    def check_status(
        self,
        response: httpx.Response | Iterable[httpx.Response] | None = None,
        data: Any | Tuple[Any, ...] | None = None,
        adapter: TypeAdapter | None = None,
        *,
        expect_err: ErrDetail | None = None,
        expect_status: int | None = None,
        request_handler_data: ResolvableRequestHandlerDatas | None = None,
    ) -> Tuple[
        Tuple[RequestHandlerData, ...],
        AssertionError | None,
    ]:
        request_handler_datas = self.resolve(
            response,
            data,
            adapter,
            expect_err=expect_err,
            expect_status=expect_status,
            request_handler_data=request_handler_data,
        )

        err = None
        if err_table := self.compile(self.collect(request_handler_datas)):
            with self.console.capture() as capture:
                self.console.print(err_table)

            err = AssertionError(capture.get())

        return request_handler_datas, err

    def compile(self, err_tables: Tuple[Table, ...]) -> Table | None:
        if len(err_tables) == 0:
            return None

        tt = Table(show_header=False, title="Errors")
        tt.add_column()
        for err_table in err_tables:
            tt.add_section()
            err_table.title = None
            tt.add_row(err_table)

        return tt
